#!/usr/bin/env python
# group_member.py
#
# Copyright (C) 2008 Veselin Penev, http://bitdust.io
#
# This file (group_member.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com


"""
.. module:: group_member
.. role:: red

BitDust group_member() Automat

EVENTS:
    * :red:`broker-connect-ack`
    * :red:`broker-connect-failed`
    * :red:`broker-lookup-ack`
    * :red:`broker-lookup-failed`
    * :red:`brokers-all-connected`
    * :red:`brokers-changed`
    * :red:`brokers-found`
    * :red:`brokers-mismatch`
    * :red:`brokers-not-found`
    * :red:`brokers-ping-failed`
    * :red:`brokers-read`
    * :red:`dht-read-failed`
    * :red:`init`
    * :red:`instant`
    * :red:`join`
    * :red:`leave`
    * :red:`message-in`
    * :red:`message-pushed`
    * :red:`push-message`
    * :red:`push-message-failed`
    * :red:`queue-in-sync`
    * :red:`queue-is-ahead`
    * :red:`queue-read-failed`
    * :red:`reconnect`
    * :red:`replace-active-broker`
    * :red:`shutdown`
    * :red:`top-broker-ping-failed`
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

import os
import re

#------------------------------------------------------------------------------

from twisted.internet import reactor  # @UnresolvedImport
from twisted.internet.defer import Deferred
from twisted.python.failure import Failure

#------------------------------------------------------------------------------

from logs import lg

from automats import automat

from lib import utime
from lib import packetid
from lib import jsn
from lib import strng
from lib import serialization

from main import events
from main import config

from crypt import my_keys

from dht import dht_relations

from contacts import identitycache

from p2p import commands
from p2p import p2p_service
from p2p import lookup
from p2p import propagate
from p2p import p2p_service_seeker

from transport import packet_out

from stream import message

from storage import archive_reader

from access import groups

from userid import identity
from userid import global_id
from userid import id_url
from userid import my_id

#------------------------------------------------------------------------------

CRITICAL_PUSH_MESSAGE_FAILS = None
MAX_BUFFERED_MESSAGES = 10
MAX_CONNECT_LOOKUP_ATTEMPTS = 5

_ActiveGroupMembers = {}
_ActiveGroupMembersByIDURL = {}

#------------------------------------------------------------------------------

def register_group_member(A):
    """
    """
    global _ActiveGroupMembers
    global _ActiveGroupMembersByIDURL
    if _Debug:
        lg.args(_DebugLevel, instance=repr(A))
    if A.group_key_id in _ActiveGroupMembers:
        raise Exception('group_member already exist')
    _ActiveGroupMembers[A.group_key_id] = A
    if id_url.is_not_in(A.group_creator_idurl, _ActiveGroupMembersByIDURL):
        _ActiveGroupMembersByIDURL[A.group_creator_idurl] = []
    _ActiveGroupMembersByIDURL[A.group_creator_idurl].append(A)
    if _Debug:
        lg.args(_DebugLevel, group_creator_idurl=A.group_creator_idurl, group_key_id=A.group_key_id)


def unregister_group_member(A):
    """
    """
    global _ActiveGroupMembers
    global _ActiveGroupMembersByIDURL
    if _Debug:
        lg.args(_DebugLevel, instance=repr(A))
    if id_url.is_not_in(A.group_creator_idurl, _ActiveGroupMembersByIDURL):
        lg.warn('for given customer idurl %r did not found active group members list' % A.group_creator_idurl)
    else:
        if A in _ActiveGroupMembersByIDURL[A.group_creator_idurl]:
            _ActiveGroupMembersByIDURL[A.group_creator_idurl].remove(A)
            if _Debug:
                lg.args(_DebugLevel, group_creator_idurl=A.group_creator_idurl, group_key_id=A.group_key_id)
        else:
            lg.warn('group_member() instance not found for customer %r' % A.group_creator_idurl)
    return _ActiveGroupMembers.pop(A.group_key_id, None)

#------------------------------------------------------------------------------

def list_active_group_members():
    """
    """
    global _ActiveGroupMembers
    return list(_ActiveGroupMembers.keys())


def get_active_group_member(group_key_id):
    """
    """
    global _ActiveGroupMembers
    if group_key_id not in _ActiveGroupMembers:
        return None
    return _ActiveGroupMembers[group_key_id]


def find_active_group_members(group_creator_idurl):
    """
    """
    global _ActiveGroupMembersByIDURL
    result = []
    for automat_index in _ActiveGroupMembersByIDURL.values():
        A = automat.by_index(automat_index, None)
        if not A:
            continue
        if A.group_creator_idurl == group_creator_idurl:
            result.append(A)
    return result

#------------------------------------------------------------------------------

def restart_active_group_member(group_key_id, use_dht_cache=False):
    if _Debug:
        lg.args(_DebugLevel, group_key_id=group_key_id, use_dht_cache=use_dht_cache)
    existing_group_member = get_active_group_member(group_key_id)
    if not existing_group_member:
        lg.err('did not found active group member %r' % group_key_id)
        return None
    result = Deferred()
    existing_index = existing_group_member.index
    existing_publish_events = existing_group_member.publish_events
    existing_group_member.automat('shutdown')
    existing_group_member = None
    del existing_group_member
    new_group_member = []

    def _on_group_member_state_changed(oldstate, newstate, event_string, *args, **kwargs):
        if _Debug:
            lg.args(_DebugLevel, oldstate=oldstate, newstate=newstate, event_string=event_string)
        if newstate == 'IN_SYNC!' and oldstate != newstate:
            new_group_member[0].removeStateChangedCallback(_on_group_member_state_changed)
            result.callback(new_group_member[0].to_json())
        if newstate == 'DISCONNECTED' and oldstate != newstate:
            new_group_member[0].removeStateChangedCallback(_on_group_member_state_changed)
            result.callback(new_group_member[0].to_json())
        return None

    def _do_start_new_group_member():
        new_group_member.append(GroupMember(
            group_key_id=group_key_id,
            use_dht_cache=use_dht_cache,
            publish_events=existing_publish_events,
        ))
        new_index = new_group_member[0].index
        new_group_member[0].automat('init')
        new_group_member[0].automat('join')
        new_group_member[0].addStateChangedCallback(_on_group_member_state_changed)
        if _Debug:
            lg.args(_DebugLevel, group_key_id=group_key_id, existing=existing_index, new=new_index)

    reactor.callLater(0, _do_start_new_group_member)  # @UndefinedVariable
    return result

#------------------------------------------------------------------------------

def rotate_active_group_memeber(old_group_key_id, new_group_key_id):
    global _ActiveGroupMembers
    A_old = get_active_group_member(old_group_key_id)
    if not A_old:
        return False
    A_new = get_active_group_member(new_group_key_id)
    if A_new and A_new in _ActiveGroupMembers:
        lg.err('it seems group %r already rotated, but older copy also exists at the moment: %r' % (A_new, A_old, ))
        return False
    del A_new  # just my paranoia
    unregister_group_member(A_old)
    A_old.update_group_key_id(new_group_key_id)
    register_group_member(A_old)
    restart_active_group_member(new_group_key_id)
    return True

#------------------------------------------------------------------------------

def start_group_members():

    def _start():
        started = 0
        for group_key_id, group_info in groups.active_groups().items():
            if not group_key_id:
                continue
            if not my_keys.is_key_registered(group_key_id):
                lg.err('can not start GroupMember because key %r is not registered' % group_key_id)
                continue
            if group_key_id.startswith('person'):
                # TODO: temporarily disabled
                continue
            if not group_info['active']:
                continue
            if not id_url.is_cached(global_id.glob2idurl(group_key_id, as_field=False)):
                continue
            existing_group_member = get_active_group_member(group_key_id)
            if not existing_group_member:
                existing_group_member = GroupMember(group_key_id)
                existing_group_member.automat('init')
            if existing_group_member.state in ['DHT_READ?', 'BROKERS?', 'QUEUE?', 'IN_SYNC!', ]:
                continue
            existing_group_member.automat('join')
            started += 1

    def _on_cached(result):
        if _Debug:
            lg.args(_DebugLevel, result=type(result))
        # a small delay to make sure received idurls were refreshed and rotated locally
        reactor.callLater(.5, _start)  # @UndefinedVariable
        return None

    idurls_to_be_updated = []
    for group_key_id, _ in groups.active_groups().items():
        if not group_key_id:
            continue
        if not my_keys.is_key_registered(group_key_id):
            continue
        if group_key_id.startswith('person'):
            # TODO: temporarily disabled
            continue
        creator_idurl = global_id.glob2idurl(group_key_id)
        if id_url.is_the_same(creator_idurl, my_id.getIDURL()):
            continue
        idurls_to_be_updated.append(creator_idurl)

    if _Debug:
        lg.args(_DebugLevel, active_groups=len(groups.active_groups()), idurls_to_be_updated=len(idurls_to_be_updated))
    d = propagate.fetch(idurls_to_be_updated, refresh_cache=True)
    d.addBoth(_on_cached)
    return True


def shutdown_group_members():
    stopped = 0
    for group_key_id in groups.active_groups().keys():
        existing_group_member = get_active_group_member(group_key_id)
        if not existing_group_member:
            continue
        existing_group_member.automat('shutdown')
        stopped += 1
    return stopped

#------------------------------------------------------------------------------

class GroupMember(automat.Automat):
    """
    This class implements all the functionality of ``group_member()`` state machine.
    """

    def __init__(
        self,
        group_key_id,
        member_idurl=None,
        use_dht_cache=True,
        debug_level=_DebugLevel,
        log_events=_Debug,
        log_transitions=_Debug,
        publish_events=False,
        **kwargs
    ):
        """
        Builds `group_member()` state machine.
        """
        self.member_idurl = member_idurl or my_id.getIDURL()
        self.member_id = self.member_idurl.to_id()
        self.group_key_id = group_key_id
        self.group_glob_id = global_id.ParseGlobalID(self.group_key_id)
        self.group_queue_alias = self.group_glob_id['key_alias']
        self.group_creator_id = self.group_glob_id['customer']
        self.group_creator_idurl = self.group_glob_id['idurl']
        self.member_sender_id = global_id.MakeGlobalID(idurl=self.member_idurl, key_alias=self.group_queue_alias)
        self.active_broker_id = None
        self.active_queue_id = None
        self.dead_broker_id = None
        self.connected_brokers = {}
        self.latest_dht_brokers = None
        self.last_sequence_id = groups.get_last_sequence_id(self.group_key_id)
        self.outgoing_messages = {}
        self.outgoing_counter = 0
        self.dht_read_use_cache = use_dht_cache
        self.buffered_messages = {}
        self.recorded_messages = []
        self.group_brokers_updated = False
        super(GroupMember, self).__init__(
            name="group_member_%s$%s" % (self.group_queue_alias[:10], self.group_creator_id),
            state="AT_STARTUP",
            debug_level=debug_level,
            log_events=log_events,
            log_transitions=log_transitions,
            publish_events=publish_events,
            **kwargs
        )

    def __repr__(self):
        connected_brokers_short = ['+' if (self.connected_brokers or {}).get(p) else ' ' for p in range(groups.REQUIRED_BROKERS_COUNT)]
        return '%s[%s](%s)' % (self.id, ''.join(connected_brokers_short), self.state)

    def update_group_key_id(self, new_group_key_id):
        if _Debug:
            lg.args(_DebugLevel, old=self.group_key_id, new=new_group_key_id)
        self.group_key_id = new_group_key_id
        self.group_glob_id = global_id.ParseGlobalID(self.group_key_id)
        self.group_queue_alias = self.group_glob_id['key_alias']
        self.group_creator_id = self.group_glob_id['customer']
        self.group_creator_idurl = self.group_glob_id['idurl']
        self.member_sender_id = global_id.MakeGlobalID(idurl=self.member_idurl, key_alias=self.group_queue_alias)

    def to_json(self):
        j = super().to_json()
        j.update({
            'active': groups.is_group_active(self.group_key_id),
            'member_id': self.member_id,
            'group_key_id': self.group_key_id,
            'alias': self.group_glob_id['key_alias'],
            'label': my_keys.get_label(self.group_key_id),
            'creator': self.group_creator_id,
            'active_broker_id': self.active_broker_id,
            'active_queue_id': self.active_queue_id,
            'connected_brokers': {p: id_url.idurl_to_id(b) for p, b in self.connected_brokers.items()},
            'last_sequence_id': self.last_sequence_id,
            'archive_folder_path': groups.get_archive_folder_path(self.group_key_id),
        })
        return j

    def init(self):
        """
        Method to initialize additional variables and flags
        at creation phase of `group_member()` machine.
        """

    def register(self):
        """
        """
        automat_index = automat.Automat.register(self)
        register_group_member(self)
        return automat_index

    def unregister(self):
        """
        """
        unregister_group_member(self)
        return automat.Automat.unregister(self)

    def state_changed(self, oldstate, newstate, event, *args, **kwargs):
        """
        This method intended to catch the moment when automat's state was changed.
        """
        if newstate == 'QUEUE?':
            self.automat('instant')
        if _Debug:
            lg.out(_DebugLevel - 2, '%s : [%s]->[%s]' % (self.name, oldstate, newstate))
        if newstate == 'IN_SYNC!':
            lg.info('group synchronized : %s' % self.group_key_id)
            events.send('group-synchronized', data=dict(
                group_key_id=self.group_key_id,
                old_state=oldstate,
                new_state=newstate,
            ))
        if newstate not in ['DISCONNECTED', 'IN_SYNC!', 'CLOSED', ] and oldstate in ['DISCONNECTED', 'IN_SYNC!', ]:
            lg.info('group connecting : %s' % self.group_key_id)
            events.send('group-connecting', data=dict(
                group_key_id=self.group_key_id,
                old_state=oldstate,
                new_state=newstate,
            ))
        if newstate == 'DISCONNECTED' and oldstate != 'AT_STARTUP':
            lg.info('group disconnected : %s' % self.group_key_id)
            events.send('group-disconnected', data=dict(
                group_key_id=self.group_key_id,
                old_state=oldstate,
                new_state=newstate,
            ))

    def A(self, event, *args, **kwargs):
        """
        The state machine code, generated using `visio2python <http://bitdust.io/visio2python/>`_ tool.
        """
        #---AT_STARTUP---
        if self.state == 'AT_STARTUP':
            if event == 'init':
                self.state = 'DISCONNECTED'
                self.SyncedUp=False
                self.doInit(*args, **kwargs)
        #---DISCONNECTED---
        elif self.state == 'DISCONNECTED':
            if event == 'leave' or event == 'shutdown':
                self.state = 'CLOSED'
                self.doDeactivate(event, *args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'join' or ( event == 'brokers-changed' and self.isActive(*args, **kwargs) ) or ( event == 'instant' and self.isActive(*args, **kwargs) and self.isDeadBroker(*args, **kwargs) ):
                self.state = 'DHT_READ?'
                self.doActivate(*args, **kwargs)
                self.doDHTReadBrokers(event, *args, **kwargs)
        #---DHT_READ?---
        elif self.state == 'DHT_READ?':
            if event == 'dht-read-failed':
                self.state = 'DISCONNECTED'
                self.SyncedUp=False
                self.doDisconnected(event, *args, **kwargs)
            elif event == 'shutdown' or event == 'leave':
                self.state = 'CLOSED'
                self.doDeactivate(event, *args, **kwargs)
                self.doCancelService(event, *args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'message-in':
                self.doRecord(*args, **kwargs)
            elif event == 'queue-in-sync':
                self.SyncedUp=True
            elif event == 'brokers-changed' or event == 'brokers-read' or event == 'brokers-found' or event == 'brokers-not-found':
                self.state = 'BROKERS?'
                self.doConnectSingleBroker(event, *args, **kwargs)
        #---BROKERS?---
        elif self.state == 'BROKERS?':
            if event == 'shutdown' or event == 'leave':
                self.state = 'CLOSED'
                self.doDeactivate(event, *args, **kwargs)
                self.doCancelService(event, *args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'message-in':
                self.doRecord(*args, **kwargs)
            elif event == 'queue-in-sync':
                self.SyncedUp=True
            elif event == 'brokers-all-connected':
                self.state = 'QUEUE?'
                self.doRememberBrokers(*args, **kwargs)
                self.doProcess(*args, **kwargs)
                self.doReadQueue(*args, **kwargs)
            elif event == 'broker-lookup-ack' or event == 'broker-connect-ack':
                self.doPingBrokers(*args, **kwargs)
            elif event == 'top-broker-ping-failed' or event == 'brokers-ping-failed' or event == 'broker-lookup-failed' or event == 'broker-connect-failed':
                self.state = 'DISCONNECTED'
            elif event == 'brokers-mismatch':
                self.doConnectSingleBroker(event, *args, **kwargs)
        #---QUEUE?---
        elif self.state == 'QUEUE?':
            if event == 'queue-read-failed':
                self.state = 'DISCONNECTED'
                self.SyncedUp=False
                self.doMarkDeadBroker(event, *args, **kwargs)
                self.doDisconnected(event, *args, **kwargs)
            elif event == 'message-in':
                self.doRecord(*args, **kwargs)
                self.doProcess(*args, **kwargs)
            elif event == 'shutdown' or event == 'leave':
                self.state = 'CLOSED'
                self.doDeactivate(event, *args, **kwargs)
                self.doCancelService(event, *args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'queue-is-ahead':
                self.doReadArchive(*args, **kwargs)
            elif event == 'queue-in-sync' or ( event == 'instant' and self.SyncedUp ):
                self.state = 'IN_SYNC!'
                self.SyncedUp=True
                self.doPushPendingMessages(*args, **kwargs)
                self.doConnected(*args, **kwargs)
            elif event == 'reconnect' or event == 'brokers-changed' or event == 'push-message-failed' or event == 'replace-active-broker':
                self.state = 'DHT_READ?'
                self.SyncedUp=False
                self.doMarkDeadBroker(event, *args, **kwargs)
                self.doDHTReadBrokers(event, *args, **kwargs)
        #---IN_SYNC!---
        elif self.state == 'IN_SYNC!':
            if event == 'push-message':
                self.doPublish(*args, **kwargs)
            elif event == 'shutdown' or event == 'leave':
                self.state = 'CLOSED'
                self.doDeactivate(event, *args, **kwargs)
                self.doCancelService(event, *args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'message-pushed':
                self.doNotifyMessageAccepted(*args, **kwargs)
            elif event == 'reconnect' or event == 'brokers-changed' or event == 'push-message-failed' or event == 'replace-active-broker':
                self.state = 'DHT_READ?'
                self.SyncedUp=False
                self.doMarkDeadBroker(event, *args, **kwargs)
                self.doDHTReadBrokers(event, *args, **kwargs)
            elif event == 'message-in':
                self.doRecord(*args, **kwargs)
                self.doProcess(*args, **kwargs)
        #---CLOSED---
        elif self.state == 'CLOSED':
            pass
        return None

    def isDeadBroker(self, *args, **kwargs):
        """
        Condition method.
        """
        if self.dead_broker_id and self.dead_broker_id == self.active_broker_id:
            return True
        return None in args[0]

    def isActive(self, *args, **kwargs):
        """
        Condition method.
        """
        return groups.is_group_active(self.group_key_id)

    def doInit(self, *args, **kwargs):
        """
        Action method.
        """
        if not groups.is_group_exist(self.group_key_id):
            groups.set_group_info(self.group_key_id)
            groups.save_group_info(self.group_key_id)
        message.consume_messages(
            consumer_callback_id=self.name,
            callback=self._do_read_queue_messages,
            direction='incoming',
            message_types=['queue_message', ],
        )
        events.add_subscriber(self._on_group_brokers_updated, 'group-brokers-updated')

    def doActivate(self, *args, **kwargs):
        """
        Action method.
        """
        groups.set_group_active(self.group_key_id, True)
        groups.save_group_info(self.group_key_id)

    def doDHTReadBrokers(self, event, *args, **kwargs):
        """
        Action method.
        """
        self.latest_dht_brokers = None
        if event == 'brokers-changed':
            self.dht_read_use_cache = True
            known_brokers = {}
            for pos, broker_id in enumerate(groups.known_brokers(self.group_creator_id)):
                if broker_id:
                    known_brokers[pos] = global_id.glob2idurl(broker_id)
            if _Debug:
                lg.args(_DebugLevel, e=event, known_brokers=known_brokers)
            self.automat('brokers-read', known_brokers=known_brokers)
            return
        if event == 'reconnect':
            cooperated_brokers = kwargs.get('cooperated_brokers')
            if cooperated_brokers:
                if _Debug:
                    lg.args(_DebugLevel, e=event, cooperated_brokers=cooperated_brokers)
                self.dht_read_use_cache = True
                self.automat('brokers-read', known_brokers=cooperated_brokers)
                return
        if event in ['reconnect', 'push-message-failed', 'replace-active-broker', 'broker-position-mismatch', 'top-broker-connect-failed', ]:
            self.dht_read_use_cache = False
        if _Debug:
            lg.args(_DebugLevel, e=event, use_cache=self.dht_read_use_cache)
        result = dht_relations.read_customer_message_brokers(
            self.group_creator_idurl,
            positions=list(range(groups.REQUIRED_BROKERS_COUNT)),
            use_cache=self.dht_read_use_cache,
        )
        # TODO: add more validations of dht_result
        result.addCallback(self._on_read_customer_message_brokers)
        if _Debug:
            result.addErrback(lg.errback, debug=_Debug, debug_level=_DebugLevel, method='group_member.doDHTReadBrokers')
        result.addErrback(lambda err: self.automat('dht-read-failed', err))

    def doConnectSingleBroker(self, event, *args, **kwargs):
        """
        Action method.
        """
        if _Debug:
            lg.args(_DebugLevel, e=event, kw=kwargs)
        target_broker = None
        if event == 'brokers-changed':
            target_broker = self._do_detect_target_broker(available_brokers=kwargs['connected_brokers'])
        elif event == 'brokers-read':
            target_broker = self._do_detect_target_broker(available_brokers=kwargs['known_brokers'])
        elif event == 'brokers-mismatch':
            groups.clear_brokers(self.group_creator_id)
            if 'dht_brokers' in kwargs:
                target_broker = self._do_detect_target_broker(available_brokers=kwargs['dht_brokers'])
            elif 'cooperated_brokers' in kwargs:
                target_broker = self._do_detect_target_broker(available_brokers=kwargs['cooperated_brokers'])
            else:
                target_broker = self._do_detect_target_broker()
        else:
            groups.clear_brokers(self.group_creator_id)
            target_broker = self._do_detect_target_broker(dht_brokers=kwargs['dht_brokers'])
        self._do_connect_target_broker(target_broker['broker_pos'], target_broker['broker_idurl'])

    def doPingBrokers(self, *args, **kwargs):
        """
        Action method.
        """
        cooperated_brokers = kwargs.get('cooperated_brokers') or {}
        archive_folder_path = kwargs.get('archive_folder_path')
        if _Debug:
            lg.args(_DebugLevel, cooperated=cooperated_brokers, archive_folder_path=archive_folder_path)
        idurl_list = [cooperated_brokers.get(i) for i in range(groups.REQUIRED_BROKERS_COUNT)]
        propagate.ping_nodes(
            idurl_list=idurl_list,
            timeout=15,
            channel='ping_broker',
            keep_alive=False,
        ).addBoth(self._on_brokers_ping_result, cooperated_brokers=cooperated_brokers, archive_folder_path=archive_folder_path)

    def doRememberBrokers(self, *args, **kwargs):
        """
        Action method.
        """
        self._do_remember_brokers(*args, **kwargs)

    def doReadQueue(self, *args, **kwargs):
        """
        Action method.
        """
        message_ack_timeout = config.conf().getInt('services/private-groups/message-ack-timeout')
        if message_ack_timeout:
            message_ack_timeout *= 2
        result = message.send_message(
            json_data={
                'msg_type': 'queue_message',
                'action': 'consume',
                'created': utime.get_sec1970(),
                'last_sequence_id': self.last_sequence_id,
                'queue_id': self.active_queue_id,
                'consumer_id': self.member_id,
            },
            recipient_global_id=self.active_broker_id,
            packet_id=packetid.MakeQueueMessagePacketID(self.active_queue_id, packetid.UniqueID()),
            message_ack_timeout=message_ack_timeout,
            skip_handshake=True,
            fire_callbacks=False,
        )
        if _Debug:
            result.addErrback(lg.errback, debug=_Debug, debug_level=_DebugLevel, method='group_member.doReadQueue')
        result.addErrback(self._on_read_queue_failed)

    def doReadArchive(self, *args, **kwargs):
        """
        Action method.
        """
        latest_known_sequence_id = kwargs.get('latest_known_sequence_id')
        received_messages = kwargs.get('received_messages')
        result_defer = Deferred()
        result_defer.addCallback(self._on_read_archive_success, received_messages)
        result_defer.addErrback(self._on_read_archive_failed, received_messages, latest_known_sequence_id)
        ar = archive_reader.ArchiveReader()
        ar.automat(
            'start',
            queue_id=self.active_queue_id,
            start_sequence_id=self.last_sequence_id + 1,
            end_sequence_id=latest_known_sequence_id,
            archive_folder_path=groups.get_archive_folder_path(self.group_key_id),
            result_defer=result_defer,
        )

    def doRecord(self, *args, **kwargs):
        """
        Action method.
        """
        self.recorded_messages.append(kwargs)

    def doProcess(self, *args, **kwargs):
        """
        Action method.
        """
        while self.recorded_messages:
            recorded_kwargs = self.recorded_messages.pop(0)
            message.push_group_message(**recorded_kwargs)
            self.automat('message-consumed', recorded_kwargs)

    def doPublish(self, *args, **kwargs):
        """
        Action method.
        """
        self._do_send_message_to_broker(
            json_payload=kwargs['json_payload'],
            outgoing_counter=None,
            packet_id=None,
        )

    def doPushPendingMessages(self, *args, **kwargs):
        """
        Action method.
        """
        if self.outgoing_messages:
            for outgoing_counter in sorted(list(self.outgoing_messages.keys())):
                self._do_send_message_to_broker(
                    json_payload=None,
                    outgoing_counter=outgoing_counter,
                    packet_id=None,
                )

    def doNotifyMessageAccepted(self, *args, **kwargs):
        """
        Action method.
        """

    def doMarkDeadBroker(self, event, *args, **kwargs):
        """
        Action method.
        """
        self.group_brokers_updated = False
        if event == 'top-broker-failed':
            self.dead_broker_id = None
            if self.latest_dht_brokers is not None:
                for broker_info in self.latest_dht_brokers:
                    if broker_info['position'] == 0 and broker_info.get('idurl'):
                        self.dead_broker_id = global_id.idurl2glob(broker_info['idurl'])
        elif event == 'replace-active-broker':
            self.dead_broker_id = args[0]
        elif event == 'brokers-changed':
            self.dead_broker_id = None
        else:
            if event == 'reconnect':
                self.dead_broker_id = None
            else:
                self.dead_broker_id = self.active_broker_id
        self.dht_read_use_cache = False
        for outgoing_counter in self.outgoing_messages.keys():
            self.outgoing_messages[outgoing_counter]['attempts'] = 0
        if _Debug:
            lg.args(_DebugLevel, dead_broker_id=self.dead_broker_id)

    def doConnected(self, *args, **kwargs):
        """
        Action method.
        """
        self.dht_read_use_cache = True
        if self.group_brokers_updated:
            events.send('group-brokers-updated', data=dict(
                group_creator_id=self.group_creator_id,
                group_key_id=self.group_key_id,
                member_id=self.member_id,
                connected_brokers=self.connected_brokers,
            ))
            self.group_brokers_updated = False

    def doDisconnected(self, *args, **kwargs):
        """
        Action method.
        """
        self.group_brokers_updated = False
        self.dht_read_use_cache = False

    def doDeactivate(self, event, *args, **kwargs):
        """
        Action method.
        """
        self.group_brokers_updated = False
        if event == 'leave':
            groups.set_group_active(self.group_key_id, False)
            groups.save_group_info(self.group_key_id)
            if kwargs.get('erase_key', False):
                if my_keys.is_key_registered(self.group_key_id):
                    my_keys.erase_key(self.group_key_id)
                else:
                    lg.warn('key %r not registered, can not be erased' % self.group_key_id)
        else:
            groups.save_group_info(self.group_key_id)

    def doCancelService(self, event, *args, **kwargs):
        """
        Action method.
        """
        if event == 'leave':
            for broker_idurl in self.connected_brokers.values():
                if not broker_idurl:
                    continue
                p2p_service.SendCancelService(
                    remote_idurl=broker_idurl,
                    service_name='service_message_broker',
                    json_payload=self._do_prepare_service_request_params(broker_idurl, action='queue-disconnect'),
                )

    def doDestroyMe(self, *args, **kwargs):
        """
        Remove all references to the state machine object to destroy it.
        """
        events.remove_subscriber(self._on_group_brokers_updated, 'group-brokers-updated')
        message.clear_consumer_callbacks(self.name)
        self.destroy()
        self.member_idurl = None
        self.member_id = None
        self.member_sender_id = None
        self.group_glob_id = None
        self.group_queue_alias = None
        self.group_creator_id = None
        self.group_creator_idurl = None
        self.active_broker_id = None
        self.active_queue_id = None
        self.dead_broker_id = None
        self.connected_brokers = None
        self.latest_dht_brokers = None
        self.outgoing_messages = None
        self.outgoing_counter = None
        self.buffered_messages = None
        self.recorded_messages = None
        self.group_brokers_updated = None

    #------------------------------------------------------------------------------

    def _do_detect_target_broker(self, available_brokers=None, dht_brokers=None):
        if _Debug:
            lg.args(_DebugLevel, dead=self.dead_broker_id, available=len(available_brokers or []), dht=len(dht_brokers or []))
        last_pos = groups.REQUIRED_BROKERS_COUNT - 1
        if self.dead_broker_id:
            if not self.active_broker_id or self.active_broker_id == self.dead_broker_id:
                return {
                    'broker_idurl': None,
                    'broker_pos': last_pos,
                }
        if available_brokers:
            if last_pos in available_brokers:
                return {
                    'broker_idurl': available_brokers[last_pos],
                    'broker_pos': last_pos,
                }
        if dht_brokers:
            for dht_broker in dht_brokers:
                if dht_broker['position'] == last_pos and dht_broker['broker_idurl']:
                    return {
                        'broker_idurl': dht_broker['broker_idurl'],
                        'broker_pos': last_pos,
                    }
        return {
            'broker_idurl': None,
            'broker_pos': last_pos,
        }

    def _do_connect_target_broker(self, broker_pos, broker_idurl):
        if broker_idurl:
            self._do_request_service_one_broker(broker_pos, broker_idurl)
        else:
            self._do_lookup_one_broker(broker_pos, skip_brokers=[])

    def _do_request_service_one_broker(self, broker_pos, broker_idurl):
        if _Debug:
            lg.args(_DebugLevel, broker_pos=broker_pos, broker_idurl=broker_idurl)
        result = p2p_service_seeker.connect_known_node(
            remote_idurl=broker_idurl,
            service_name='service_message_broker',
            service_params=lambda idurl: self._do_prepare_service_request_params(idurl, broker_pos),
            request_service_timeout=config.conf().getInt('services/private-groups/broker-connect-timeout', 120),
            attempts=1,
        )
        result.addCallback(self._on_broker_connected, broker_pos)
        if _Debug:
            result.addErrback(lg.errback, debug=_Debug, debug_level=_DebugLevel, method='group_member._do_request_service_one_broker')
        result.addErrback(self._on_broker_connect_failed, broker_pos)

    def _do_lookup_one_broker(self, broker_pos, skip_brokers):
        if _Debug:
            lg.args(_DebugLevel, broker_pos=broker_pos, skip_brokers=skip_brokers)
        exclude_brokers = set()
        for known_broker_id in groups.known_brokers(self.group_creator_id):
            if known_broker_id:
                exclude_brokers.add(id_url.to_bin(global_id.glob2idurl(known_broker_id, as_field=False)))
        for connected_broker_idurl in self.connected_brokers.values():
            exclude_brokers.add(id_url.to_bin(connected_broker_idurl))
        for skip_idurl in skip_brokers:
            if skip_idurl:
                exclude_brokers.add(id_url.to_bin(skip_idurl))
        if self.dead_broker_id:
            exclude_brokers.add(id_url.to_bin(global_id.glob2idurl(self.dead_broker_id, as_field=False)))
        preferred_brokers = []
        preferred_brokers_raw = config.conf().getData('services/private-groups/preferred-brokers').strip()
        if preferred_brokers_raw:
            preferred_brokers_list = re.split('\n|,|;| ', preferred_brokers_raw)
            preferred_brokers.extend(preferred_brokers_list)
            preferred_brokers = id_url.to_bin_list(preferred_brokers)
        if preferred_brokers:
            preferred_brokers = [x for x in preferred_brokers if x not in exclude_brokers]
        if _Debug:
            lg.args(_DebugLevel, exclude_brokers=exclude_brokers, preferred_brokers=preferred_brokers)
        if preferred_brokers:
            preferred_broker_idurl = id_url.field(preferred_brokers[0])
            if preferred_broker_idurl and id_url.is_not_in(preferred_broker_idurl, exclude_brokers, as_field=False):
                result = p2p_service_seeker.connect_known_node(
                    remote_idurl=preferred_broker_idurl,
                    service_name='service_message_broker',
                    service_params=lambda idurl: self._do_prepare_service_request_params(idurl, broker_pos),
                    request_service_timeout=config.conf().getInt('services/private-groups/broker-connect-timeout', 120),
                    exclude_nodes=list(exclude_brokers),
                    attempts=1,
                )
                result.addCallback(self._on_broker_hired, broker_pos)
                if _Debug:
                    result.addErrback(lg.errback, debug=_Debug, debug_level=_DebugLevel, method='group_member._do_lookup_one_broker')
                result.addErrback(self._on_broker_connect_failed, broker_pos)
                return result
        result = p2p_service_seeker.connect_random_node(
            lookup_method=lookup.random_message_broker,
            service_name='service_message_broker',
            service_params=lambda idurl: self._do_prepare_service_request_params(idurl, broker_pos),
            request_service_timeout=config.conf().getInt('services/private-groups/broker-connect-timeout', 120),
            exclude_nodes=list(exclude_brokers),
            attempts=2,
        )
        result.addCallback(self._on_broker_hired, broker_pos)
        if _Debug:
            result.addErrback(lg.errback, debug=_Debug, debug_level=_DebugLevel, method='group_member._do_lookup_one_broker')
        result.addErrback(self._on_broker_lookup_failed, broker_pos)
        return result

    def _do_read_queue_messages(self, json_messages):
        if not json_messages:
            return True
        if _Debug:
            lg.args(_DebugLevel, active_queue_id=self.active_queue_id, active_broker_id=self.active_broker_id, json_messages=len(json_messages))
        latest_known_sequence_id = -1
        received_group_messages = []
        packets_to_ack = {}
        cooperated_brokers = {}
        to_be_reconnected = False
        found_group_ids = set()
        found_broker_ids = set()
        for json_message in json_messages:
            try:
                msg_type = json_message.get('msg_type', '') or json_message.get('type', '')  # TODO: need to cleanup that later
                msg_direction = json_message['dir']
                packet_id = json_message['packet_id']
                owner_idurl = json_message['owner_idurl']
                msg_data = json_message['data']
                msg_action = msg_data.get('action', 'read')
            except:
                lg.exc()
                continue
            if msg_direction != 'incoming':
                continue
            if msg_type != 'queue_message':
                continue
            if msg_action != 'read':
                continue
            if 'last_sequence_id' not in msg_data:
                continue
            try:
                chunk_last_sequence_id = int(msg_data['last_sequence_id'])
                list_messages = msg_data['items']
            except:
                lg.exc(msg_data)
                continue
            incoming_queue_id = packetid.SplitQueueMessagePacketID(packet_id)[0]
            incoming_group_alias, incoming_group_creator_id, incoming_broker_id = global_id.SplitGlobalQueueID(incoming_queue_id)
            incoming_group_creator_id = global_id.glob2idurl(incoming_group_creator_id).to_id()
            incoming_group_key_id = global_id.MakeGlobalKeyID(incoming_group_alias, incoming_group_creator_id)
            found_group_ids.add(incoming_group_key_id)
            found_broker_ids.add(incoming_broker_id)
            if incoming_group_key_id != self.group_key_id:
                if _Debug:
                    lg.dbg(_DebugLevel, 'skip message based on packet_id for %r : %r' % (self.group_key_id, incoming_group_key_id, ))
                continue
            if chunk_last_sequence_id > latest_known_sequence_id:
                latest_known_sequence_id = chunk_last_sequence_id
            for one_message in list_messages:
                if one_message['sequence_id'] > latest_known_sequence_id:
                    lg.warn('invalid item sequence_id %d   vs.  last_sequence_id %d known' % (
                        one_message['sequence_id'], latest_known_sequence_id))
                    continue
                group_message_object = message.GroupMessage.deserialize(one_message['payload'])
                if group_message_object is None:
                    lg.err("GroupMessage deserialize failed, can not extract message from payload of %d bytes" % len(one_message['payload']))
                    continue
                try:
                    decrypted_message = group_message_object.decrypt()
                    json_message = serialization.BytesToDict(
                        decrypted_message,
                        unpack_types=True,
                        encoding='utf-8',
                    )
                except:
                    lg.exc()
                    continue
                received_group_messages.append(dict(
                    json_message=json_message,
                    direction='incoming',
                    group_key_id=self.group_key_id,
                    producer_id=one_message['producer_id'],
                    sequence_id=one_message['sequence_id'],
                ))
                if owner_idurl:
                    packets_to_ack[packet_id] = owner_idurl
        for packet_id, owner_idurl in packets_to_ack.items():
            p2p_service.SendAckNoRequest(owner_idurl, packet_id)
            received_queue_id, _, _ = packet_id.rpartition('_') 
            _, received_broker_id = global_id.SplitGlobalQueueID(received_queue_id, split_queue_alias=False)
            if self.active_broker_id and received_broker_id != self.active_broker_id:
                if not to_be_reconnected:
                    to_be_reconnected = True
                    lg.warn('received message from broker %r which is different from my active broker %r' % (
                        received_broker_id, self.active_broker_id, ))
        if received_group_messages and len(found_broker_ids) > 0 and self.active_broker_id not in found_broker_ids:
            to_be_reconnected = True
            lg.warn('active broker is %r but incoming message received from another broker %r in %r' % (
                self.active_broker_id, list(found_broker_ids), self, ))
        packets_to_ack.clear()
        if to_be_reconnected:
            cooperated_brokers = json_message.get('cooperated_brokers', {}) or {}
            cooperated_brokers = {int(k): id_url.field(v) for k,v in cooperated_brokers.items()}
            if _Debug:
                lg.args(_DebugLevel, cooperated_brokers=cooperated_brokers)
        if not received_group_messages:
            if self.group_key_id not in found_group_ids:
                if _Debug:
                    lg.dbg(_DebugLevel, 'no messages for %r found in the incoming stream' % self.active_queue_id)
                return True
            if latest_known_sequence_id < self.last_sequence_id:
                lg.warn('found queue latest sequence %d is behind of my current position %d' % (latest_known_sequence_id, self.last_sequence_id, ))
                self.automat('queue-in-sync')
                if to_be_reconnected:
                    if _Debug:
                        lg.dbg(_DebugLevel, 'going to reconnect %r' % self)
                    reactor.callLater(0.01, self.automat, 'reconnect', cooperated_brokers=cooperated_brokers)  # @UndefinedVariable
                return True
            if latest_known_sequence_id > self.last_sequence_id:
                lg.warn('nothing received, but found queue latest sequence %d is ahead of my current position %d, need to read messages from archive' % (
                    latest_known_sequence_id, self.last_sequence_id, ))
                self.automat('queue-is-ahead', latest_known_sequence_id=latest_known_sequence_id, received_messages=received_group_messages, )
                if to_be_reconnected:
                    if _Debug:
                        lg.dbg(_DebugLevel, 'going to reconnect %r' % self)
                    reactor.callLater(0.01, self.automat, 'reconnect', cooperated_brokers=cooperated_brokers)  # @UndefinedVariable
                return True
            self.last_sequence_id = latest_known_sequence_id
            groups.set_last_sequence_id(self.group_key_id, latest_known_sequence_id)
            groups.save_group_info(self.group_key_id)
            if _Debug:
                lg.dbg(_DebugLevel, 'no new messages, queue in sync, latest_known_sequence_id=%d' % latest_known_sequence_id)
            self.automat('queue-in-sync')
            if to_be_reconnected:
                if _Debug:
                    lg.dbg(_DebugLevel, 'going to reconnect %r' % self)
                reactor.callLater(0.01, self.automat, 'reconnect', cooperated_brokers=cooperated_brokers)  # @UndefinedVariable
            return True
        received_group_messages.sort(key=lambda m: m['sequence_id'])
        ret = self._do_process_group_messages(received_group_messages, latest_known_sequence_id)
        if to_be_reconnected:
            if _Debug:
                lg.dbg(_DebugLevel, 'going to reconnect %r' % self)
            self.automat('reconnect', cooperated_brokers=cooperated_brokers)
        return ret

    def _do_process_group_messages(self, received_group_messages, latest_known_sequence_id):
        if _Debug:
            lg.args(_DebugLevel, received_group_messages=len(received_group_messages), buffered_messages=len(self.buffered_messages), latest_known_sequence_id=latest_known_sequence_id)
        newly_processed = 0
        for new_message in received_group_messages:
            new_sequence_id = new_message['sequence_id']
            if new_sequence_id in self.buffered_messages:
                lg.warn('message %d already buffered' % new_sequence_id)
                continue
            self.buffered_messages[new_sequence_id] = new_message
        buffered_sequence_ids = sorted(self.buffered_messages.keys())
        for new_sequence_id in buffered_sequence_ids:
            if self.last_sequence_id + 1 == new_sequence_id:
                inp_message = self.buffered_messages.pop(new_sequence_id)
                self.last_sequence_id = new_sequence_id
                newly_processed += 1
                groups.set_last_sequence_id(self.group_key_id, self.last_sequence_id)
                groups.save_group_info(self.group_key_id)
                lg.info('new message consumed in %r, last_sequence_id incremented to %d' % (self.group_key_id, self.last_sequence_id, ))
                self.automat('message-in', **inp_message)
        if len(self.buffered_messages) > MAX_BUFFERED_MESSAGES:
            raise Exception('message sequence is broken by message broker %s, currently %d buffered messages' % (self.active_broker_id, len(self.buffered_messages), ))
        if _Debug:
            lg.dbg(_DebugLevel, 'my_last_sequence_id=%d  newly_processed=%d  buffered_messages=%d' % (
                self.last_sequence_id, newly_processed, len(self.buffered_messages), ))
        if not newly_processed or newly_processed != len(received_group_messages):
            if latest_known_sequence_id > self.last_sequence_id:
                lg.warn('found queue latest sequence %d is ahead of my current position %d, need to read messages from archive' % (
                    latest_known_sequence_id, self.last_sequence_id, ))
                self.automat('queue-is-ahead', latest_known_sequence_id=latest_known_sequence_id, received_messages=received_group_messages, )
                return True
        if _Debug:
            lg.dbg(_DebugLevel, 'processed all messages, queue in sync, last_sequence_id=%d' % self.last_sequence_id)
        self.automat('queue-in-sync')
        return True

    def _do_send_message_to_broker(self, json_payload=None, outgoing_counter=None, packet_id=None):
        global CRITICAL_PUSH_MESSAGE_FAILS
        if CRITICAL_PUSH_MESSAGE_FAILS is None:
            CRITICAL_PUSH_MESSAGE_FAILS = int(os.environ.get('BITDUST_CRITICAL_PUSH_MESSAGE_FAILS', 2))
        if packet_id is None:
            packet_id = packetid.UniqueID()
        if _Debug:
            lg.args(_DebugLevel, json_payload=json_payload, outgoing_counter=outgoing_counter, packet_id=packet_id)
        require_handshake = False
        if outgoing_counter is None:
            self.outgoing_counter += 1
            outgoing_counter = self.outgoing_counter
            self.outgoing_messages[outgoing_counter] = {
                'attempts': 0,
                'last_attempt': None,
                'payload': json_payload,
            }
        else:
            if self.outgoing_messages[outgoing_counter]['attempts'] > CRITICAL_PUSH_MESSAGE_FAILS:
                lg.err('failed sending message to broker %r after %d attempts' % (
                    self.active_broker_id, self.outgoing_messages[outgoing_counter]['attempts'], ))
                self.outgoing_messages[outgoing_counter]['attempts'] = 0
                self.outgoing_messages[outgoing_counter]['last_attempt'] = None
                self.automat('push-message-failed')
                return
            if self.outgoing_messages[outgoing_counter]['last_attempt'] is not None:
                if utime.get_sec1970() - self.outgoing_messages[outgoing_counter]['last_attempt'] < config.conf().getInt('services/private-groups/message-ack-timeout'):
                    lg.warn('pending message %d already made attempt to send recently' % outgoing_counter)
                    return
            self.outgoing_messages[outgoing_counter]['attempts'] += 1
            self.outgoing_messages[outgoing_counter]['last_attempt'] = utime.get_sec1970()
            json_payload = self.outgoing_messages[outgoing_counter]['payload']
            if self.outgoing_messages[outgoing_counter]['attempts'] >= 1:
                require_handshake = True
            lg.warn('re-trying sending message to broker %r   counter=%d attempts=%d packet_id=%s' % (
                self.active_broker_id, outgoing_counter, self.outgoing_messages[outgoing_counter]['attempts'], packet_id, ))
        raw_payload = serialization.DictToBytes(
            json_payload,
            pack_types=True,
            encoding='utf-8',
        )
        try:
            private_message_object = message.GroupMessage(
                recipient=self.group_key_id,
                sender=self.member_sender_id,
            )
            private_message_object.encrypt(raw_payload)
        except:
            lg.exc()
            raise Exception('message encryption failed')
        encrypted_payload = private_message_object.serialize()
        d = message.send_message(
            json_data={
                'msg_type': 'queue_message',
                'action': 'produce',
                'created': utime.get_sec1970(),
                'payload': encrypted_payload,
                'queue_id': self.active_queue_id,
                'producer_id': self.member_id,
                'brokers': self.connected_brokers,
            },
            recipient_global_id=self.active_broker_id,
            packet_id=packetid.MakeQueueMessagePacketID(self.active_queue_id, packet_id),
            message_ack_timeout=config.conf().getInt('services/private-groups/message-ack-timeout'),
            skip_handshake=True,
            fire_callbacks=False,
            require_handshake=require_handshake,
        )
        if _Debug:
            d.addErrback(lg.errback, debug=_Debug, debug_level=_DebugLevel, method='group_member._do_send_message_to_broker')
        d.addCallback(self._on_message_to_broker_sent, outgoing_counter, packet_id)
        d.addErrback(self._on_message_to_broker_failed, outgoing_counter, packet_id)

    def _do_prepare_service_request_params(self, possible_broker_idurl, desired_broker_position=-1, action='queue-connect'):
        if _Debug:
            lg.args(_DebugLevel, broker=possible_broker_idurl, pos=desired_broker_position, action=action, owner=self.group_creator_id, )
        group_key_info = {}
        # if not my_keys.is_key_registered(self.group_key_id):
        #     lg.warn('group key %r was not registered, checking all registered keys' % self.group_key_id)
        #     my_keys.check_rename_my_keys(prefix=self.group_key_id.split('@')[0])
        if not my_keys.is_key_registered(self.group_key_id):
            lg.err('closing group_member %r because key %r is not registered' % (self, self.group_key_id, ))
            lg.exc('group key %r is not registered' % self.group_key_id)
            self.automat('shutdown')
            return
        try:
            group_key_info = my_keys.get_key_info(
                key_id=self.group_key_id,
                include_private=False,
                include_signature=True,
                generate_signature=True,
                include_label=False,
            )
        except:
            lg.exc()
            self.automat('shutdown')
            return
        service_request_params = {
            'action': action,
            'queue_id': None,
            'consumer_id': self.member_id,
            'producer_id': self.member_id,
            'group_key': group_key_info,
            'archive_folder_path': groups.get_archive_folder_path(self.group_key_id),
            'last_sequence_id': self.last_sequence_id,
            'known_brokers': self.connected_brokers,
        }
        if action == 'queue-disconnect':
            service_request_params['queue_id'] = self.active_queue_id
        if desired_broker_position >= 0:
            service_request_params['position'] = desired_broker_position
        if _Debug:
            lg.args(_DebugLevel, action=action, last_sequence_id=service_request_params['last_sequence_id'],
                    archive_folder_path=service_request_params['archive_folder_path'])
        return service_request_params

    def _do_remember_brokers(self, *args, **kwargs):
        if _Debug:
            lg.args(_DebugLevel, args=args, kwargs=kwargs)
        cooperated_brokers = kwargs.get('cooperated_brokers') or {}
        current_connected_brokers = dict(self.connected_brokers or {})
        archive_folder_path = kwargs.get('archive_folder_path')
        brokers_changed = False
        self.active_broker_id = None
        self.active_queue_id = None
        self.connected_brokers = {}
        for position in range(groups.REQUIRED_BROKERS_COUNT):
            broker_idurl = cooperated_brokers.get(position)
            if broker_idurl:
                broker_idurl = id_url.field(broker_idurl)
                broker_id = global_id.idurl2glob(broker_idurl)
                groups.set_broker(self.group_creator_id, broker_id, position)
                self.connected_brokers[position] = broker_idurl
                if position == 0:
                    self.dead_broker_id = None
                    self.active_broker_id = broker_id
                    self.active_queue_id = global_id.MakeGlobalQueueID(
                        queue_alias=self.group_queue_alias,
                        owner_id=self.group_creator_id,
                        supplier_id=self.active_broker_id,
                    )
            else:
                groups.clear_broker(self.group_creator_id, position)
            if current_connected_brokers.get(position) != self.connected_brokers.get(position):
                brokers_changed = True
        if self.active_broker_id is None:
            raise Exception('active broker was not connected')
        if archive_folder_path:
            current_path = groups.get_archive_folder_path(self.group_key_id)
            if current_path and current_path != archive_folder_path:
                lg.warn('for %r overwriting existing archive_folder_path %r with new value %r' % (
                    self.group_key_id, current_path, archive_folder_path, ))
            else:
                if not current_path:
                    lg.info('recognized archive folder path for %r from broker response: %r' % (
                        self.group_key_id, archive_folder_path, ))
            groups.set_archive_folder_path(self.group_key_id, archive_folder_path)
        groups.save_group_info(self.group_key_id)
        if brokers_changed:
            self.group_brokers_updated = True

    #------------------------------------------------------------------------------

    def _on_read_customer_message_brokers(self, brokers_info_list):
        if _Debug:
            lg.args(_DebugLevel, brokers=len(brokers_info_list))
        if not brokers_info_list:
            self.dht_read_use_cache = False
            self.automat('brokers-not-found', dht_brokers=[])
            return
        self.latest_dht_brokers = brokers_info_list
        self.automat('brokers-found', dht_brokers=brokers_info_list)

    def _on_broker_connected(self, response_info, broker_pos, *args, **kwargs):
        if _Debug:
            lg.args(_DebugLevel, resp=response_info, broker_pos=broker_pos, args=args, kwargs=kwargs)
        try:
            # skip leading "accepted:" marker
            cooperated_brokers = jsn.loads(strng.to_text(response_info[0].Payload)[9:])
            archive_folder_path = strng.to_text(cooperated_brokers.pop('archive_folder_path', None))
            cooperated_brokers = {int(k): id_url.field(v) for k,v in cooperated_brokers.items()}
        except:
            lg.exc()
            self.automat('broker-connect-failed')
            return
        self.automat('broker-connect-ack', cooperated_brokers=cooperated_brokers, archive_folder_path=archive_folder_path)

    def _on_broker_connect_failed(self, err, broker_pos, *args, **kwargs):
        if _Debug:
            lg.args(_DebugLevel, err=err, broker_pos=broker_pos, args=args, kwargs=kwargs)
        if isinstance(err, Failure):
            if _Debug:
                lg.args(_DebugLevel, args=err.value.args)
            mismatch_info = {}
            try:
                evt, a, kw = err.value.args
                if a and a[0]:
                    if isinstance(a[0], packet_out.PacketOut):
                        resp_payload = strng.to_text(a[0].Payload)
                    else:
                        resp_payload = strng.to_text(a[0][0].Payload)
                    if resp_payload.startswith('identity:'):
                        xml_src = resp_payload[9:]
                        new_ident = identity.identity(xmlsrc=xml_src)
                        if new_ident.isCorrect() and new_ident.Valid():
                            if identitycache.UpdateAfterChecking(new_ident.getIDURL(), xml_src):
                                reactor.callLater(0.2, self.automat, 'brokers-mismatch')  # @UndefinedVariable
                                return
                    if resp_payload.startswith('mismatch:'):
                        mismatch_info = jsn.loads(resp_payload[9:])
                        if 'dht_brokers' in mismatch_info:
                            mismatch_info['dht_brokers'] = {int(k): id_url.field(v) for k,v in mismatch_info['dht_brokers'].items()}
                        if 'cooperated_brokers' in mismatch_info:
                            mismatch_info['cooperated_brokers'] = {int(k): id_url.field(v) for k,v in mismatch_info['cooperated_brokers'].items()}
            except:
                lg.exc()
                self.automat('broker-connect-failed', err)
                return
            if mismatch_info:
                lg.warn('broker request mismatch at position %d: %r' % (broker_pos, mismatch_info, ))
                self.automat('brokers-mismatch', **mismatch_info)
                return
        lg.err('failed connecting to broker at position %d : %r' % (broker_pos, err, ))
        self.automat('broker-connect-failed', err)

    def _on_broker_hired(self, response_info, broker_pos, *args, **kwargs):
        if _Debug:
            lg.args(_DebugLevel, resp=response_info, broker_pos=broker_pos, args=args, kwargs=kwargs)
        try:
            # skip leading "accepted:" marker
            cooperated_brokers = jsn.loads(strng.to_text(response_info[0].Payload)[9:])
            archive_folder_path = strng.to_text(cooperated_brokers.pop('archive_folder_path', None))
            cooperated_brokers = {int(k): id_url.field(v) for k,v in cooperated_brokers.items()}
        except:
            lg.exc()
            self.automat('broker-lookup-failed')
            return
        self.automat('broker-lookup-ack', cooperated_brokers=cooperated_brokers, archive_folder_path=archive_folder_path)

    def _on_broker_lookup_failed(self, err, broker_pos, *args, **kwargs):
        if _Debug:
            lg.args(_DebugLevel, err=err, broker_pos=broker_pos)
        if isinstance(err, Failure):
            if _Debug:
                lg.args(_DebugLevel, args=err.value.args)
            mismatch_info = {}
            try:
                evt, a, kw = err.value.args
                if a and a[0]:
                    if isinstance(a[0], packet_out.PacketOut):
                        resp_payload = strng.to_text(a[0].Payload)
                    else:
                        resp_payload = strng.to_text(a[0][0].Payload)
                    if resp_payload.startswith('mismatch:'):
                        mismatch_info = jsn.loads(resp_payload[9:])
                        if 'dht_brokers' in mismatch_info:
                            mismatch_info['dht_brokers'] = {int(k): id_url.field(v) for k,v in mismatch_info['dht_brokers'].items()}
                        if 'cooperated_brokers' in mismatch_info:
                            mismatch_info['cooperated_brokers'] = {int(k): id_url.field(v) for k,v in mismatch_info['cooperated_brokers'].items()}
            except:
                lg.exc()
                self.automat('broker-lookup-failed', err)
                return
            if mismatch_info:
                self.automat('brokers-mismatch', **mismatch_info)
                return
        self.automat('broker-lookup-failed', err)

    def _on_brokers_ping_result(self, ping_results, **kwargs):
        if _Debug:
            lg.args(_DebugLevel, ping_results=ping_results, kwargs=kwargs)
        try:
            all_results = [r[0] for r in ping_results]
            top_broker_result = ping_results[0][0]
        except:
            lg.exc()
            top_broker_result = False
        if not top_broker_result:
            self.automat('top-broker-ping-failed')
            return
        if True not in all_results:
            self.automat('brokers-ping-failed')
            return
        self.automat('brokers-all-connected', **kwargs)

    def _on_read_queue_failed(self, err):
        if _Debug:
            lg.args(_DebugLevel, err=err)
        if isinstance(err, Failure):
            if _Debug:
                lg.args(_DebugLevel, args=err.value.args)
            try:
                resp_payload = strng.to_text(err.value.args[0].Payload)
                if _Debug:
                    lg.dbg(_DebugLevel, resp_payload)
                if resp_payload.startswith('identity:'):
                    xml_src = resp_payload[9:]
                    new_ident = identity.identity(xmlsrc=xml_src)
                    if new_ident.isCorrect() and new_ident.Valid():
                        if identitycache.UpdateAfterChecking(new_ident.getIDURL(), xml_src):
                            reactor.callLater(0.5, self.automat, 'reconnect')  # @UndefinedVariable
                            return
            except:
                lg.exc()
        self.automat('queue-read-failed', err)

    def _on_message_to_broker_sent(self, response_packet, outgoing_counter, packet_id):
        if _Debug:
            lg.args(_DebugLevel, response_packet=response_packet, outgoing_counter=outgoing_counter)
        if outgoing_counter not in self.outgoing_messages:
            raise Exception('outgoing message with counter %d not found' % outgoing_counter)
        if response_packet and response_packet.Command == commands.Ack():
            self.outgoing_messages.pop(outgoing_counter)
            self.automat('message-pushed', outgoing_counter=outgoing_counter)
            return
        self._do_send_message_to_broker(json_payload=None, outgoing_counter=outgoing_counter, packet_id=None)

    def _on_message_to_broker_failed(self, err, outgoing_counter, packet_id):
        if _Debug:
            lg.args(_DebugLevel, err=err, outgoing_counter=outgoing_counter, packet_id=packet_id)
        self.outgoing_messages[outgoing_counter]['last_attempt'] = None
        self._do_send_message_to_broker(json_payload=None, outgoing_counter=outgoing_counter, packet_id=None)

    def _on_read_archive_success(self, archive_messages, received_messages):
        if _Debug:
            lg.args(_DebugLevel, archive_messages=len(archive_messages), received_messages=len(received_messages))
        received_group_messages = []
        latest_known_sequence_id = -1
        for archive_message in archive_messages:
            if archive_message['sequence_id'] > latest_known_sequence_id:
                latest_known_sequence_id = archive_message['sequence_id']
        for archive_message in archive_messages:
            received_group_messages.append(dict(
                json_message=archive_message['payload'],
                direction='incoming',
                group_key_id=self.group_key_id,
                producer_id=archive_message['producer_id'],
                sequence_id=archive_message['sequence_id'],
            ))
        for received_message in received_messages:
            received_group_messages.append(dict(
                json_message=received_message['json_message'],
                direction='incoming',
                group_key_id=self.group_key_id,
                producer_id=received_message['producer_id'],
                sequence_id=received_message['sequence_id'],
            ))
            if received_message['sequence_id'] > latest_known_sequence_id:
                latest_known_sequence_id = received_message['sequence_id']
        self._do_process_group_messages(received_group_messages, latest_known_sequence_id)

    def _on_read_archive_failed(self, err, received_messages, latest_known_sequence_id):
        lg.warn('received %d recent messages but read archived messages failed with: %r' % (len(received_messages), err, ))
        self.last_sequence_id = latest_known_sequence_id
        groups.set_last_sequence_id(self.group_key_id, latest_known_sequence_id)
        groups.save_group_info(self.group_key_id)
        self.automat('queue-in-sync')
        return None

    def _on_group_brokers_updated(self, evt):
        d = evt.data
        if d.get('group_creator_id') != self.group_creator_id:
            return
        if d.get('group_key_id') == self.group_key_id:
            return
        changed = False
        for pos, broker_idurl in d['connected_brokers'].items():
            if pos not in self.connected_brokers:
                changed = True
                break
            if self.connected_brokers[pos] != broker_idurl:
                changed = True
                break
        if _Debug:
            lg.args(_DebugLevel, changed=changed, this=repr(self), connected_brokers=d['connected_brokers'])
        if changed:
            reactor.callLater(0.01, self.automat, 'brokers-changed', connected_brokers=d['connected_brokers'])  # @UndefinedVariable



#!/usr/bin/python
# message_database.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (message_database.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#

"""
..

module:: message_database
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
from __future__ import print_function

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 6

#------------------------------------------------------------------------------

import os
import json
import sqlite3

#------------------------------------------------------------------------------

if __name__ == '__main__':
    import sys
    import os.path as _p
    sys.path.insert(0, _p.abspath(_p.join(_p.dirname(_p.abspath(sys.argv[0])), '..')))

#------------------------------------------------------------------------------

from logs import lg

from lib import utime
from lib import strng

from main import settings

from crypt import my_keys

from userid import my_id

#------------------------------------------------------------------------------

_HistoryDB = None
_HistoryCursor = None

#------------------------------------------------------------------------------

MESSAGE_TYPES = {
    'message': 1,
    'private_message': 2,
    'group_message': 3,
    'personal_message': 4,
}

MESSAGE_TYPE_CODES = {
    1: 'message',
    2: 'private_message',
    3: 'group_message',
    4: 'personal_message',
}

#------------------------------------------------------------------------------

def init(filepath=None):
    """
    """
    global _HistoryDB
    global _HistoryCursor

    if not filepath:
        filepath = settings.ChatMessagesHistoryDatabaseFile()

    if _Debug:
        lg.args(_DebugLevel, filepath=filepath)

    sqlite3.register_adapter(dict, adapt_json)
    sqlite3.register_adapter(list, adapt_json)
    sqlite3.register_adapter(tuple, adapt_json)
    sqlite3.register_converter('JSON', convert_json)

    if not os.path.isfile(filepath):
        _HistoryDB = sqlite3.connect(filepath, timeout=1)
        _HistoryDB.execute('PRAGMA case_sensitive_like = 1;')
        _HistoryCursor = _HistoryDB.cursor()

        _HistoryCursor.execute('''CREATE TABLE IF NOT EXISTS "history" (
            "sender_local_key_id" INTEGER,
            "sender_id" TEXT,
            "recipient_local_key_id" INTEGER,
            "recipient_id" TEXT,
            "direction" INTEGER,
            "payload_type" INTEGER,
            "payload_time" INTEGER,
            "payload_message_id" TEXT,
            "payload_body" JSON)''')
        _HistoryCursor.execute('CREATE INDEX "sender local key id" on history(sender_local_key_id)')
        _HistoryCursor.execute('CREATE INDEX "recipient local key id" on history(recipient_local_key_id)')

        _HistoryCursor.execute('''CREATE TABLE IF NOT EXISTS "conversations" (
            "conversation_id" TEXT,
            "payload_type" INTEGER,
            "started_time" INTEGER,
            "last_updated_time" INTEGER,
            "last_message_id" TEXT)''')
        _HistoryCursor.execute('CREATE INDEX "conversation id" on conversations(conversation_id)')

        _HistoryCursor.execute('''CREATE TABLE IF NOT EXISTS "keys" (
            "key_id" TEXT,
            "local_key_id" INTEGER,
            "public_key" TEXT)''')
        _HistoryCursor.execute('CREATE INDEX "key id" on keys(key_id)')
        _HistoryCursor.execute('CREATE INDEX "local key id" on keys(local_key_id)')
        _HistoryCursor.execute('CREATE INDEX "public key" on keys(public_key)')

        _HistoryDB.commit()
        _HistoryDB.close()

    _HistoryDB = sqlite3.connect(filepath, timeout=1)
    _HistoryDB.text_factory = str
    _HistoryDB.execute('PRAGMA case_sensitive_like = 1;')
    _HistoryDB.commit()
    _HistoryCursor = _HistoryDB.cursor()


def shutdown():
    """
    """
    global _HistoryDB
    global _HistoryCursor
    if _Debug:
        lg.dbg(_DebugLevel, '')

    _HistoryDB.commit()
    _HistoryDB.close()
    _HistoryDB = None
    _HistoryCursor = None

#------------------------------------------------------------------------------

def db(instance='current'):
    global _HistoryDB
    return _HistoryDB


def cur(instance='current'):
    global _HistoryCursor
    return _HistoryCursor

#------------------------------------------------------------------------------

def adapt_json(data):
    return (json.dumps(data, sort_keys=True)).encode()

def convert_json(blob):
    return json.loads(blob.decode())

#------------------------------------------------------------------------------

def build_json_message(data, message_id, message_time=None, sender=None, recipient=None, message_type=None, direction=None):
    """
    """
    if not sender:
        sender = my_id.getGlobalID(key_alias='master')
    if not recipient:
        recipient = my_id.getGlobalID(key_alias='master')
    if direction is None:
        if message_type in ['private_message', None, ]:
            direction = 'out' if sender == my_id.getGlobalID(key_alias='master') else 'in'
        else:
            direction = 'in'
    else:
        direction = direction.replace('incoming', 'in').replace('outgoing', 'out')
    new_json = {
        "payload": {
            "msg_type": message_type or "message",
            "message_id": strng.to_text(message_id),
            "time": message_time or utime.utcnow_to_sec1970(),
            "data": data,
        },
        'sender': {
            'glob_id': sender,
        },
        'recipient': {
            'glob_id': recipient,
        },
        'direction': direction,
    }
    return new_json

#------------------------------------------------------------------------------

def insert_message(message_json):
    """
    Writes JSON message to the message database.
    """
    try:
        sender_glob_id = message_json['sender']['glob_id']
        recipient_glob_id = message_json['recipient']['glob_id']
        direction = message_json['direction']
        payload_type = MESSAGE_TYPES.get(message_json['payload']['msg_type'], 1)
        payload_time = message_json['payload']['time']
        payload_message_id = message_json['payload']['message_id']
        payload_body =  message_json['payload']['data']
    except:
        lg.exc()
        return False
    if _Debug:
        lg.args(_DebugLevel, sender=sender_glob_id, recipient=recipient_glob_id, typ=payload_type, message_id=payload_message_id)
    recipient_local_key_id = my_keys.get_local_key_id(recipient_glob_id)
    if payload_type in [3, 4, ]:
        sender_local_key_id = recipient_local_key_id
    else:
        sender_local_key_id = my_keys.get_local_key_id(sender_glob_id)
    if sender_local_key_id is None or recipient_local_key_id is None:
        lg.err('failed to store message because local_key_id is not found, sender=%r recipient=%r' % (
            sender_local_key_id, recipient_local_key_id, ))
        return False
    cur().execute('''INSERT INTO history (
            sender_local_key_id,
            sender_id,
            recipient_local_key_id,
            recipient_id,
            direction,
            payload_type,
            payload_time,
            payload_message_id,
            payload_body
        ) VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)''', (
        sender_local_key_id,
        sender_glob_id,
        recipient_local_key_id,
        recipient_glob_id,
        0 if direction == 'in' else 1,
        payload_type,
        payload_time,
        payload_message_id,
        payload_body,
    ))
    db().commit()
    update_conversation(sender_local_key_id, recipient_local_key_id, payload_type, payload_time, payload_message_id)
    return True


def update_conversation(sender_local_key_id, recipient_local_key_id, payload_type, payload_time, payload_message_id):
    conversation_id = None
    if payload_type in [3, 4, ]:
        conversation_id = '{}&{}'.format(recipient_local_key_id, recipient_local_key_id)
    elif payload_type == 2:
        if recipient_local_key_id < sender_local_key_id:
            conversation_id = '{}&{}'.format(recipient_local_key_id, sender_local_key_id)
        else:
            conversation_id = '{}&{}'.format(sender_local_key_id, recipient_local_key_id)
    else:
        lg.err('unexpected message type: %r' % payload_type)
        return
    if conversation_id is None:
        lg.err('failed to update conversation, local_key_id was not found')
        return
    sql = 'SELECT * FROM conversations WHERE conversation_id=?'
    params = [conversation_id, ]
    found_conversation = list(cur().execute(sql, params))
    if found_conversation:
        sql = 'UPDATE conversations SET last_updated_time=?, last_message_id=? WHERE conversation_id=?'
        params = [payload_time, payload_message_id, conversation_id, ]
    else:
        sql = 'INSERT INTO conversations (conversation_id, payload_type, started_time, last_updated_time, last_message_id) VALUES (?, ?, ?, ?, ?)'
        params = [conversation_id, payload_type, payload_time, payload_time, payload_message_id, ]
    if _Debug:
        lg.args(_DebugLevel, conversation_id=conversation_id, found_conversation=len(found_conversation), params=params)
    cur().execute(sql, params)
    db().commit()
    return True


def query_messages(sender_id=None, recipient_id=None, bidirectional=True, order_by_time=True, message_types=[], offset=None, limit=None):
    """
    """
    sql = 'SELECT * FROM history'
    q = ''
    params = []
    if bidirectional and sender_id and recipient_id:
        recipient_local_key_id = my_keys.get_local_key_id(recipient_id)
        sender_local_key_id = my_keys.get_local_key_id(sender_id)
        if recipient_local_key_id is None or sender_local_key_id is None:
            lg.warn('local_key_id was not found, recipient_local_key_id=%r sender_local_key_id=%r' % (recipient_local_key_id, sender_local_key_id, ))
            return []
        q += ' sender_local_key_id IN (?, ?) AND recipient_local_key_id IN (?, ?)'
        params += [sender_local_key_id, recipient_local_key_id, sender_local_key_id, recipient_local_key_id, ]
    else:
        if sender_id:
            sender_local_key_id = my_keys.get_local_key_id(sender_id)
            if sender_local_key_id is None:
                lg.warn('local_key_id was not found for sender %r' % sender_id)
                return []
            q += ' sender_local_key_id=?'
            params += [sender_local_key_id, ]
        if recipient_id:
            recipient_local_key_id = my_keys.get_local_key_id(recipient_id)
            if recipient_local_key_id is None:
                lg.warn('local_key_id was not found for recipient %r' % recipient_id)
                return []
            q += ' recipient_local_key_id=?'
            params += [recipient_local_key_id, ]
    if message_types:
        if params:
            q += ' AND payload_type IN (%s)' % (','.join(['?', ] * len(message_types)))
        else:
            q += ' payload_type IN (%s)' % (','.join(['?', ] * len(message_types)))
        params.extend([MESSAGE_TYPES.get(mt, 1) for mt in message_types])
    if q:
        sql += ' WHERE %s' % q
    if order_by_time:
        sql += ' ORDER BY payload_time DESC'
    if limit is not None:
        sql += ' LIMIT ?'
        params += [limit, ]
    if offset is not None:
        sql += ' OFFSET ?'
        params += [offset, ]
    if _Debug:
        lg.args(_DebugLevel, sql=sql, params=params)
    results = []
    local_key_ids = {}
    for row in cur().execute(sql, params):
        sender_local_key_id = row[0]
        sender_id_recorded = row[1]
        recipient_local_key_id = row[2]
        recipient_id_recorded = row[3]
        if sender_local_key_id not in local_key_ids:
            local_key_ids[sender_local_key_id] = my_keys.get_local_key(sender_local_key_id)
        if recipient_local_key_id not in local_key_ids:
            local_key_ids[recipient_local_key_id] = my_keys.get_local_key(recipient_local_key_id)
        if not local_key_ids.get(sender_local_key_id) or not local_key_ids.get(recipient_local_key_id):
            lg.warn('unknown sender or recipient local key_id')
            continue
        json_msg = build_json_message(
            sender=sender_id_recorded,
            recipient=recipient_id_recorded,
            direction='in' if row[4] == 0 else 'out',
            message_type=MESSAGE_TYPE_CODES.get(int(row[5]), 'private_message'),
            message_time=row[6],
            message_id=row[7],
            data=json.loads(row[8]),
        )
        if order_by_time:
            results.insert(0, json_msg)
        else:
            results.append(json_msg)
    return results


def list_conversations(order_by_time=True, message_types=[], offset=None, limit=None):
    sql = 'SELECT * FROM conversations'
    q = ''
    params = []
    if message_types:
        q += ' payload_type IN (%s)' % (','.join(['?', ] * len(message_types)))
        params.extend([MESSAGE_TYPES.get(mt, 1) for mt in message_types])
    if q:
        sql += ' WHERE %s' % q
    if order_by_time:
        sql += ' ORDER BY last_updated_time DESC'
    if limit is not None:
        sql += ' LIMIT ?'
        params += [limit, ]
    if offset is not None:
        sql += ' OFFSET ?'
        params += [offset, ]
    if _Debug:
        lg.args(_DebugLevel, sql=sql, params=params)
    results = []
    for row in cur().execute(sql, params):
        results.append(dict(
            conversation_id=row[0],
            type=MESSAGE_TYPE_CODES.get(int(row[1]), 'private_message'), 
            started=row[2],
            last_updated=row[3],
            last_message_id=row[4],
        ))
    return results


#------------------------------------------------------------------------------

def update_history_with_new_local_key_id(old_id, new_id):
    sql = 'UPDATE history SET sender_local_key_id=? WHERE sender_local_key_id=?'
    params = [new_id, old_id, ]
    cur().execute(sql, params)
    db().commit()
    if _Debug:
        lg.args(_DebugLevel, sql=sql, params=params)
    sql = 'UPDATE history SET recipient_local_key_id=? WHERE recipient_local_key_id=?'
    params = [new_id, old_id, ]
    cur().execute(sql, params)
    db().commit()
    if _Debug:
        lg.args(_DebugLevel, sql=sql, params=params)


def update_conversations_with_new_local_key_id(old_id, new_id):
    sql = 'SELECT * FROM conversations'
    params = []
    modifications = {}
    for row in cur().execute(sql, params):
        conversation_id = row[0]
        new_conversation_id = conversation_id
        id1, _, id2 = conversation_id.partition('&')
        try:
            id1 = int(id1)
            id2 = int(id2)
        except:
            lg.exc()
            continue
        if id1 == old_id:
            new_conversation_id = '{}&{}'.format(new_id, id2)
        if id2 == old_id:
            new_conversation_id = '{}&{}'.format(id1, new_id)
        if conversation_id != new_conversation_id:
            modifications[conversation_id] = new_conversation_id
    if _Debug:
        lg.args(_DebugLevel, modifications=modifications)
    for old_conv_id, new_conv_id in modifications.items():
        sql = 'UPDATE conversations SET conversation_id=? WHERE conversation_id=?'
        params = [new_conv_id, old_conv_id, ]
        cur().execute(sql, params)
        db().commit()
        if _Debug:
            lg.args(_DebugLevel, sql=sql, params=params)


def rebuild_conversations():
    cur().execute('SELECT count(name) FROM sqlite_master WHERE type="table" AND name="conversations"')
    if cur().fetchone()[0]:
        return
    cur().execute('''CREATE TABLE IF NOT EXISTS "conversations" (
        "conversation_id" TEXT,
        "payload_type" INTEGER,
        "started_time" INTEGER,
        "last_updated_time" INTEGER,
        "last_message_id" TEXT)''')
    cur().execute('CREATE INDEX "conversation id" on conversations(conversation_id)')
    db().commit()
    for message_json in list(query_messages()):
        msg_typ = message_json['payload'].get('msg_type') or message_json['payload'].get('type')
        payload_type = MESSAGE_TYPES.get(msg_typ, 1)
        recipient_local_key_id = my_keys.get_local_key_id(message_json['recipient']['glob_id'])
        if payload_type in [3, 4, ]:
            sender_local_key_id = recipient_local_key_id
        else:
            sender_local_key_id = my_keys.get_local_key_id(message_json['sender']['glob_id'])
        update_conversation(
            sender_local_key_id=sender_local_key_id,
            recipient_local_key_id=recipient_local_key_id,
            payload_type=payload_type,
            payload_time=message_json['payload']['time'],
            payload_message_id=message_json['payload']['message_id'],
        )

#------------------------------------------------------------------------------

def check_create_rename_key(new_public_key, new_key_id, new_local_key_id):
    conversation_type = None
    if new_key_id.startswith('group_'):
        conversation_type = 'group_message'
    elif new_key_id.startswith('person_'):
        conversation_type = 'personal_message'
    elif new_key_id.startswith('master$'):
        conversation_type = 'private_message'
    if conversation_type is None:
        return False
    sql = 'SELECT * FROM keys WHERE public_key=?'
    params = [new_public_key, ]
    found_public_keys = list(cur().execute(sql, params))
    if found_public_keys:
        if len(found_public_keys) > 1:
            raise Exception('found multiple records for same public key: %r' % found_public_keys)
        key_id = found_public_keys[0][0]
        local_key_id = found_public_keys[0][1]
        changed = False
        if key_id != new_key_id:
            changed = True
            sql = 'UPDATE keys SET key_id=? WHERE public_key=?'
            params = [new_key_id, new_public_key, ]
            cur().execute(sql, params)
            db().commit()
            if _Debug:
                lg.args(_DebugLevel, sql=sql, params=params)
        if local_key_id != new_local_key_id:
            changed = True
            lg.warn('found new public key which is re-using already known key with different local key id: %r -> %r' % (
                local_key_id, new_local_key_id, ))
            update_history_with_new_local_key_id(local_key_id, new_local_key_id)
            update_conversations_with_new_local_key_id(local_key_id, new_local_key_id)
            sql = 'UPDATE keys SET local_key_id=? WHERE public_key=?'
            params = [new_local_key_id, new_public_key, ]
            cur().execute(sql, params)
            db().commit()
            if _Debug:
                lg.args(_DebugLevel, sql=sql, params=params)
        return changed
    # new key will be registered in the "keys" table
    # also new conversation record can be created if that was a group type
    # unlike groups, private messages suppose to record a new conversation only when a first message appears and delivered between two
    sql = 'INSERT INTO keys (key_id, local_key_id, public_key) VALUES (?, ?, ?)'
    params = [new_key_id, new_local_key_id, new_public_key, ]
    cur().execute(sql, params)
    db().commit()
    if _Debug:
        lg.args(_DebugLevel, sql=sql, params=params)
    if conversation_type in ['group_message', 'personal_message', ]:
        update_conversation(
            sender_local_key_id=new_local_key_id,
            recipient_local_key_id=new_local_key_id,
            payload_type=MESSAGE_TYPES.get(conversation_type, 1),
            payload_time=utime.utcnow_to_sec1970(),
            payload_message_id='',
        )
    return True

#------------------------------------------------------------------------------

def main():
    import pprint
    my_keys.init()
    init()
    pprint.pprint(list(query_messages(
        # sender_id='',
        recipient_id='group_96b1fa942b556e2e0dbccdcc9eeed84a$veselibro@seed.bitdust.io',
        # offset=2,
        # limit=3,
        message_types=['group_message', ]
    )))
    # pprint.pprint(list(list_conversations(
        # message_types=['group_message', ]
    # )))
    shutdown()


if __name__ == "__main__":
    lg.set_debug_level(20)
    main()



from unittest import TestCase
import os

from logs import lg

from system import bpio

from main import settings

from crypt import key

from userid import my_id



_some_priv_key = """-----BEGIN RSA PRIVATE KEY-----
MIIEowIBAAKCAQEA/ZsJKyCakqA8vO2r0CTOG0qE2l+4y1dIqh7VC0oaVkXy0Cim
s3N+NNt6vJ617H5FeMALasRf4+5nHkleTB+BKTKBxt4fLm9Mz3mkw8U+pstJfYzS
FjjG3Zjs6XC4HyMjOCTkYCdvQQTeqB9ALAJ10YpVwlJlXOaFoscwv3LkUD4XB2Q4
rh6iqYdQo7Na2vY0aAw+t/wTnLTzr/rxQuhbng7ktPv8ZUx69eHzG07gLRpqzYH7
zkvuijtesxPIFFLaSlgQPc3rILqwLP/wI6fDjtPWlnG1cJJ9e7PBq1LkQ8FJoCxa
Sx9/zrZcVuKFnCPGrc1hjqCIU9Hh3wTPxV/PIQIDAQABAoIBAAO2fZXPXf33WTj2
SpI4e2PqRVEduXAQ2AIwGNQX4T70j8A5C82Tpg7ObLYxWruICCVjJo5Oin1vpIbi
Ac66im68yTTnnyRJajynyh1jivtV3XPYatdlvJvFePNMZHtbDYa4AQ0wlJwVRFB+
LSIxJXIO94/pz9y3uLAEINWZCt31xEpIWiJYDYvOnu/SUNJ652WzmEoguzdYndam
PjoC8S4PUWslCeMiHc5XLhrZVXoNFQyqAX/zZoKLtbNQq0Z3Huguih0ID5oSFWMl
MeFLUoW2UAvo+dwkW5DJc5VyBFvAX/J+03WqvqmX0198mD8LrMDr9hGsIAyQmGud
oECFIZECgYEA/hm8SiLaMjhvtvhPDYKG7+RbEomdf2F1sSrJtyUCTCswKpjGwDnP
zcxDJuGzIRYw16GoUBWIcJfrDXzzNN6TxX9qXhd8uJ+VckEp7H1mOYLnLN5Ekjyu
XqrYzG/cSRqb1IESG3ikm8/587WxRA+4/rm+XWg1XZCOotiJrCdz6LECgYEA/4Ba
auGkYErkw2E02WMRqo2Ct2cfWPCRXy/oWSvbDxNlhFoheKr/6e9PHQ+m5iLDo3sq
3ZaPgfDaty6uI4v/rCwDGdv2s5bkI9crZ2DxNKhgJUE9No87jSSHjBMv0eylUBHA
DB5n1jyEYFHTGpy4Hvr6M6sCz/8Ehvga9J596XECgYAl4F4xytotjD4Szxaxk9hb
X/W3YK4Kc9OgUhl5ZFngUru+TcGqm7N/IMiNviz+bJlhOyaksWECL5MJEqwKIHd0
hBat6eBcgOU8/7upFdQsFHgzNvqPtd3kHKFub+otN3stBQRW9ffLhgfjLR08YP+Z
cMSQld0GkmrAmXiEIelkMQKBgQCnyIUX3ReRuHzjpQkMnJc0Vft6LvkR8eC1DoPZ
UwhmrQkkUf/a+whVejaM0gN662doCvEKVN3mqeEnHDt00nHSgZCDwcQPCU7GDn3N
RIcBWnTQ4jethX/I3y04Gj1z8KBapV1lV+4+bL1Nd05XEoWCqrP1jB6rsj8p1vH0
o8PxsQKBgFxHBaLtYSqwQoolo0mmpGCHTdfmn+bRQaqL3DQEGYULtkj+gGuqVlj6
5cnFfTRgzj81rSkJIg7ZQ+V3CvtZSjYnflN2L1IS1aT+N9uvoLSV/yzNLmHlG6nR
TjpvKeAap/ydSo6j8K97O8h4EeVKM2xfbNBVpv1mpR8AbkJf7Ioz
-----END RSA PRIVATE KEY-----"""

_some_identity_xml = """<?xml version="1.0" encoding="utf-8"?>
<identity>
  <sources>
    <source>http://127.0.0.1:8084/alice.xml</source>
  </sources>
  <contacts>
    <contact>tcp://127.0.0.1:7103</contact>
  </contacts>
  <certificates/>
  <scrubbers/>
  <postage>0</postage>
  <date>Oct 06, 2018</date>
  <version></version>
  <revision>0</revision>
  <publickey>ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAABAQD9mwkrIJqSoDy87avQJM4bSoTaX7jLV0iqHtULShpWRfLQKKazc34023q8nrXsfkV4wAtqxF/j7mceSV5MH4EpMoHG3h8ub0zPeaTDxT6my0l9jNIWOMbdmOzpcLgfIyM4JORgJ29BBN6oH0AsAnXRilXCUmVc5oWixzC/cuRQPhcHZDiuHqKph1Cjs1ra9jRoDD63/BOctPOv+vFC6FueDuS0+/xlTHr14fMbTuAtGmrNgfvOS+6KO16zE8gUUtpKWBA9zesgurAs//Ajp8OO09aWcbVwkn17s8GrUuRDwUmgLFpLH3/OtlxW4oWcI8atzWGOoIhT0eHfBM/FX88h</publickey>
  <signature>906441963064925827454594808955119786427327091644004203121255573673324339896720684156436969809331302081059171001468697744607176215971850395448358447221987030255681283469602485017439100048818715591137645334649289784704347547476744833593959051239987455656492991690929110057776187651745011046495572223232179906271885029639566994962961751420295596050170876718201976232073313442092469148257068924122205928495395843354406033087633663157320015417942886925426409641297819144861516715728591303834175552581452070473350075769438068777969720417764064190628218036180170101856467748070658740635441921977032005183554438715223073852</signature>
</identity>"""


class Test(TestCase):

    def setUp(self):
        try:
            bpio.rmdir_recursive('/tmp/.bitdust_tmp')
        except Exception:
            pass
        lg.set_debug_level(30)
        settings.init(base_dir='/tmp/.bitdust_tmp')
        self.my_current_key = None
        try:
            os.makedirs('/tmp/.bitdust_tmp/metadata/')
        except:
            pass
        fout = open('/tmp/_some_priv_key', 'w')
        fout.write(_some_priv_key)
        fout.close()
        fout = open(settings.LocalIdentityFilename(), 'w')
        fout.write(_some_identity_xml)
        fout.close()
        self.assertTrue(key.LoadMyKey(keyfilename='/tmp/_some_priv_key'))
        self.assertTrue(my_id.loadLocalIdentity())

    def tearDown(self):
        key.ForgetMyKey()
        my_id.forgetLocalIdentity()
        settings.shutdown()
        os.remove('/tmp/_some_priv_key')
        bpio.rmdir_recursive('/tmp/.bitdust_tmp')

    def test_identity_valid(self):
        from userid import identity
        some_identity = identity.identity(xmlsrc=_some_identity_xml)
        self.assertTrue(some_identity.isCorrect())
        self.assertTrue(some_identity.Valid())

    def test_identity_not_valid(self):
        from userid import identity
        _broken_identity_xml = _some_identity_xml.replace('alice', 'bob')
        broken_identity = identity.identity(xmlsrc=_broken_identity_xml)
        self.assertTrue(broken_identity.isCorrect())
        self.assertFalse(broken_identity.Valid())



#!/usr/bin/python
# service_shared_data.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (service_shared_data.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#

"""
..

module:: service_shared_data
"""

from __future__ import absolute_import
from services.local_service import LocalService


def create_service():
    return SharedDataService()


class SharedDataService(LocalService):

    service_name = 'service_shared_data'
    config_path = 'services/shared-data/enabled'

    def dependent_on(self):
        return [
            'service_my_data',
        ]

    def start(self):
        from main import events
        from transport import callback
        callback.append_inbox_callback(self._on_inbox_packet_received)
        events.add_subscriber(self._on_supplier_modified, 'supplier-modified')
        events.add_subscriber(self._on_my_list_files_refreshed, 'my-list-files-refreshed')
        events.add_subscriber(self._on_key_erased, 'key-erased')
        self._do_open_known_shares()
        return True

    def stop(self):
        from main import events
        from transport import callback
        events.remove_subscriber(self._on_key_erased, 'key-erased')
        events.remove_subscriber(self._on_my_list_files_refreshed, 'my-list-files-refreshed')
        events.remove_subscriber(self._on_supplier_modified, 'supplier-modified')
        callback.remove_inbox_callback(self._on_inbox_packet_received)
        return True

    def _on_key_erased(self, evt):
        from access import shared_access_coordinator
        active_share = shared_access_coordinator.get_active_share(evt.data['key_id'])
        if active_share:
            active_share.automat('shutdown')

    def _on_supplier_modified(self, evt):
        from access import key_ring
        from crypt import my_keys
        from userid import global_id
        from userid import my_id
        from logs import lg
        if evt.data['new_idurl']:
            my_keys_to_be_republished = []
            for key_id in my_keys.known_keys():
                if not key_id.startswith('share_'):
                    continue
                _glob_id = global_id.ParseGlobalID(key_id)
                if _glob_id['idurl'].to_bin() == my_id.getIDURL().to_bin():
                    # only send public keys of my own shares
                    my_keys_to_be_republished.append(key_id)
            for key_id in my_keys_to_be_republished:
                d = key_ring.transfer_key(key_id, trusted_idurl=evt.data['new_idurl'], include_private=False, include_signature=False)
                d.addErrback(lambda *a: lg.err('transfer key failed: %s' % str(*a)))

    def _on_my_list_files_refreshed(self, evt):
        from access import shared_access_coordinator
        from customer import supplier_connector
        from p2p import p2p_service
        from p2p import commands
        for key_id in shared_access_coordinator.list_active_shares():
            cur_share = shared_access_coordinator.get_active_share(key_id)
            if not cur_share:
                continue
            if cur_share.state != 'CONNECTED':
                continue
            for supplier_idurl in cur_share.known_suppliers_list:
                sc = supplier_connector.by_idurl(
                    supplier_idurl,
                    customer_idurl=cur_share.customer_idurl,
                )
                if sc is not None and sc.state == 'CONNECTED':
                    p2p_service.SendListFiles(
                        target_supplier=supplier_idurl,
                        customer_idurl=cur_share.customer_idurl,
                        key_id=cur_share.key_id,
                        timeout=30,
                        callbacks={
                            commands.Files(): lambda r, i: self._on_list_files_response(r, i, cur_share.customer_idurl, supplier_idurl, cur_share.key_id),
                            commands.Fail(): lambda r, i: self._on_list_files_failed(r, i, cur_share.customer_idurl, supplier_idurl, cur_share.key_id),
                        }
                    )

    def _on_list_files_response(self, response, info, customer_idurl, supplier_idurl, key_id):
        from logs import lg
        # TODO: remember the response and prevent sending ListFiles() too often

    def _on_list_files_failed(self, response, info, customer_idurl, supplier_idurl, key_id):
        from logs import lg
        from lib import strng
        from access import key_ring
        if strng.to_text(response.Payload) == 'key not registered' or strng.to_text(response.Payload) == '':
            lg.warn('supplier %r of customer %r do not possess public key %r yet, sending it now' % (
                supplier_idurl, customer_idurl, key_id, ))
            result = key_ring.transfer_key(key_id, supplier_idurl, include_private=False, include_signature=False)
            result.addCallback(lambda r: self._on_key_transfer_success(customer_idurl, supplier_idurl, key_id))
            result.addErrback(lambda err: lg.err('failed sending key %r to %r : %r' % (key_id, supplier_idurl, err, )))
        else:
            lg.err('failed requesting ListFiles() with %r for customer %r from supplier %r' % (
                key_id, customer_idurl, supplier_idurl, ))
        return None

    def _on_key_transfer_success(self, customer_idurl, supplier_idurl, key_id):
        from logs import lg
        from p2p import p2p_service
        lg.info('public key %r shared to supplier %r of customer %r, now will send ListFiles()' % (key_id, supplier_idurl, customer_idurl))
        p2p_service.SendListFiles(
            target_supplier=supplier_idurl,
            customer_idurl=customer_idurl,
            key_id=key_id,
            timeout=30,
        )

    def _on_inbox_packet_received(self, newpacket, info, status, error_message):
        from p2p import commands
        from access import key_ring
        if newpacket.Command == commands.Files():
            return key_ring.on_files_received(newpacket, info)
        return False

    def _do_open_known_shares(self):
        from crypt import my_keys
        from access import shared_access_coordinator
        from storage import backup_fs
        known_offline_shares = []
        for key_id in my_keys.known_keys():
            if not key_id.startswith('share_'):
                continue
            active_share = shared_access_coordinator.get_active_share(key_id)
            if active_share:
                continue
            known_offline_shares.append(key_id)
        to_be_opened = []
        for pathID, localPath, itemInfo in backup_fs.IterateIDs():
            if not itemInfo.key_id:
                continue
            if itemInfo.key_id in to_be_opened:
                continue
            if itemInfo.key_id not in known_offline_shares:
                continue
            to_be_opened.append(itemInfo.key_id)
        for key_id in to_be_opened:
            active_share = shared_access_coordinator.SharedAccessCoordinator(key_id, log_events=True, publish_events=False, )
            active_share.automat('restart')



#!/usr/bin/python
# backup_schedule.py
#
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (backup_schedule.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
#

"""
.. module:: backup_schedule.

A set of methods to run backups at given intervals, used to provide scheduled backups.

TODO:
need to turn on this module, it is switched off during last "big changes".
many things was changed since that moment, so need to put some work here
"""

from __future__ import absolute_import
from logs import lg
import six
from io import open

#------------------------------------------------------------------------------

_SheduledTasks = {}
_UsingGUI = False
_GuiStatusClearPageData = None
_GuiBackupUpdate = None

#------------------------------------------------------------------------------


def init(usingGui=False):
    global _UsingGUI
    _UsingGUI = usingGui
    lg.out(4, 'backup_schedule.init ')
    loop()


def shutdown():
    lg.out(4, 'backup_schedule.shutdown ')
    global _SheduledTasks
    for dirName, task in _SheduledTasks.items():
        try:
            task.cancel()
        except:
            lg.out(1, 'backup_schedule.shutdown ERROR can not stop task for ' + str(dirName))
            lg.exc()
        lg.out(4, 'backup_schedule.shutdown canceled: ' + str(dirName))


def loop():
    lg.out(8, 'backup_schedule.loop ')
    #debugWrite("in loop about to run at "+str(time.asctime(time.localtime(time.time()))))
    # run()
    #debugWrite("in loop finished run, setting callLater")
    # reactor.callLater(timeout(), loop)


# will check all Shedules each 60 minutes
def timeout():
    return 60.0 * 60.0


# def run():
#    global _SheduledTasks
#    try:
#        lg.out(8, 'backup_schedule.run ')
#        for dirName in backup_db.GetBackupDirectories():
#            lg.out(8, 'backup_schedule.run %s' % dirName)
#            if backup_db.IsBackupRunning(dirName):
#                lg.out(8, 'backup_schedule.run %s is running at the moment, skip.' % dirName)
#                continue
#
#            next_start = next(dirName)
#
#            if next_start is None:
#                # lg.out(8, '  backup_schedule.run next_start is None')
#                continue
#
#            if next_start < 0:
#                continue
#
#            now = time.time()
#            delay = next_start - now
#            lg.out(8, '  backup_schedule.run delay=%s next_start=[%s]' % (str(delay), str(time.asctime(time.localtime(next_start)))))
#            if delay > 0 and delay < timeout():
#                if _SheduledTasks.has_key(unicode(dirName)):
#                    lg.out(8, 'backup_schedule.run cancel previous task')
#                    try:
#                        _SheduledTasks[unicode(dirName)].cancel()
#                    except:
#                        pass # may have already run or already been cancelled, nothing to cancel
#                    del _SheduledTasks[unicode(dirName)]
#                _SheduledTasks[unicode(dirName)] = reactor.callLater(delay, start_backup, dirName)
#                # lg.out(8, '  backup_schedule.run will start after %s seconds' % (str(delay)))
#                # lg.out(8, '  backup_schedule.run getTime()=' + str(_SheduledTasks[unicode(dirName)].getTime()))
#
#    except:
#        lg.exc()


# def start_backup(dirName):
#    global _SheduledTasks
#    if _SheduledTasks.has_key(unicode(dirName)): # we're running now, no longer need the schedule
#        del _SheduledTasks[unicode(dirName)]
#    lg.out(6, 'backup_schedule.start_backup ' + str(dirName))
#
#    if not backup_db.CheckDirectory(dirName):
#        return
#
#    bid = misc.NewBackupID()
#    # backups.AddBackupInProcess(bid)
#    recursive_subfolders = backup_db.GetDirectorySubfoldersInclude(dirName)
#    dir_size = bpio.getDirectorySize(dirName, recursive_subfolders)
#    result = Deferred()
#    result.addCallback(backup_done)
#    result.addErrback(backup_fail)
#    dobackup.dobackup(bid, dirName, dir_size, recursive_subfolders, webcontrol.OnBackupDataPacketResult, result)
#
#    global _UsingGUI
#    global _GuiBackupUpdate
#    if _UsingGUI:
#        _GuiBackupUpdate()


# def backup_done(backupID):
#    lg.out(4, 'backup_schedule.backup_done ' + str(backupID))
#
#    global _UsingGUI
#    global _GuiBackupUpdate
#    global _GuiStatusClearPageData
#
#    if _UsingGUI:
#        _GuiStatusClearPageData(backupID)
#
#    aborted = False
#    if backupID.endswith(' abort'):
#        backupID = backupID[:-6]
#        aborted = True
#
#    backupDir = backup_db.GetDirectoryFromBackupId(backupID)
#    if backupDir == "" or backupDir is None:
#        lg.out(6, 'backup_schedule.backup_done  can not find %s  it in database' % backupID)
#        return
#
#    if aborted:
#        backupDir = backup_db.GetDirectoryFromBackupId(backupID)
#        backup_db.SetBackupStatus(backupDir, backupID, "stopped", "")
#    else:
#        backupDir = backup_db.GetDirectoryFromBackupId(backupID)
#        backup_db.SetBackupStatus(backupDir, backupID, "done", str(time.time()))
#    backup_db.Save()
#
#    # backups.RemoveBackupInProcess(backupID)
#    backup_monitor.Restart()
#
#    if aborted:
#        if _UsingGUI:
#            _GuiBackupUpdate()
#    else:
#        if _UsingGUI:
#            _GuiBackupUpdate()


# def backup_fail(backupID):
#    lg.out(4, 'backup_schedule.backup_fail ' + str(backupID))
#    global _UsingGUI
#    global _GuiBackupUpdate
#    global _GuiStatusClearPageData
#    if _UsingGUI:
#        _GuiStatusClearPageData(backupID)
#
#    backupDir = backup_db.GetDirectoryFromBackupId(backupID)
#    if backupDir == "" or backupDir is None:
#        lg.out(6, 'backup_schedule.backup_fail  can not find %s  it in database' % str(backupID))
#        return
#
#    backup_db.SetBackupStatus(backupDir, backupID, "failed", "")
#    backup_db.Save()
#    # backups.RemoveBackupInProcess(backupID)
#    backup_monitor.Restart()
#
#    if _UsingGUI:
#        _GuiBackupUpdate()


def task(dirName=None):
    global _SheduledTasks
    if dirName is None:
        return _SheduledTasks
    return _SheduledTasks.get(six.text_type(dirName), None)


def next(dirName):
    global _SheduledTasks
#    lastRunId, lastRunSize, lastRunStatus, lastRunStart, lastRunFinish = backup_db.GetLastRunInfo(dirName)
#    schedule_type, schedule_time, schedule_interval, interval_details = backup_db.GetSchedule(dirName)
    schedule = None  # backup_db.GetSchedule(dirName)
    if schedule is None:
        return None

#    lg.out(8, 'backup_schedule.next dirName=%s type=%s, time=%s, interval=%s, details=%s' % (str(dirName), schedule_type, schedule_time, schedule_interval, interval_details))
#    lg.out(8, 'backup_schedule.next %s %s' % (str(dirName), schedule))

    return schedule.next_time()

#    next_start = None
#    if schedule_type == 'none':
#        next_start = -1
#
#    elif schedule_type == 'hourly':
#        next_start = maths.shedule_next_hourly(lastRunStart, schedule_interval)
#
#    elif schedule_type == 'daily':
#        next_start = maths.shedule_next_daily(lastRunStart, schedule_interval, schedule_time)
#
#    elif schedule_type == 'weekly':
#        week_days = interval_details.split(' ')
#        week_day_numbers = []
#        week_day_names = list(calendar.day_name)
#        week_day_abbr = list(calendar.day_abbr)
#        for week_label in week_days:
#            try:
#                i = week_day_names.index(week_label)
#            except:
#                try:
#                    i = week_day_abbr.index(week_label)
#                except:
#                    continue
#            week_day_numbers.append(i)
#        next_start = maths.shedule_next_weekly(lastRunStart, schedule_interval, schedule_time, week_day_numbers)
#
#    elif schedule_type == 'monthly':
#        months_labels = interval_details.split(' ')
#        months_numbers = []
#        months_names = list(calendar.month_name)
#        months_abbr = list(calendar.month_abbr)
#        for month_label in months_labels:
#            try:
#                i = months_names.index(month_label)
#            except:
#                try:
#                    i = months_abbr.index(month_label)
#                except:
#                    continue
#            months_numbers.append(i)
#        next_start = maths.shedule_next_monthly(lastRunStart, schedule_interval, schedule_time, months_numbers)
#
#    return next_start


def debugWrite(debugtext):  # useful when debugging this module, otherwise don't use (leave False)
    if False:
        debugFile = open("scheduledebug.txt", "a")
        debugFile.write(debugtext + "\r\n")
        debugFile.close()
    else:
        lg.out(8, debugtext)


# def types(k=None):
#    d = {'0': 'none',
#         '1': 'hourly',
#         '2': 'daily',
#         '3': 'weekly',
#         '4': 'monthly',
#         '5': 'continuously'}
#    if k is None:
#        return d
#    return d.get(k, d['0'])


# def labels():
#    return {'n': 'none',
#            'h': 'hourly',
#            'd': 'daily',
#            'w': 'weekly',
#            'm': 'monthly',
#            'c': 'continuously'}


# def default():
#    return {'type': 'none',
#            'interval': '1',
#            'time': '',
#            'details': '',}


# def format():
#    return '''
# Schedule compact format:
#[mode].[interval].[time].[details]
#
# mode:
#  n-none, h-hourly, d-daily, w-weekly, m-monthly
#
# interval:
#  just a number - how often to restart the task, default is 1
#
# time:
#  [hour]:[minute]
#
# details:
#  for weeks: Mon Tue Wed Thu Fri Sat Sun
#  for months: Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec
#
# some examples:
#  n.1..                   no schedule
#  hourly.3                each 3 hours
#  daily.4.10:15.          every 4th day at 10:15
#  w.1.3:00.MonSat         every Monday and Saturday at 3:00 in the night
#  weekly.4.18:45.MonTueWedThuFriSatSun
#                          every day in each 4th week in 18:45
#  m.5.12:34.JanJul        5th Jan and 5th July at 12:34
#'''


# def unpack(s):
#    try:
#        parts = s.split('.')
#        parts += ['']*(4-len(parts))
#        (sh_type, sh_interval, sh_time, sh_details) = parts[0:4]
#        sh_type = sh_type.lower()
#        if sh_type not in ['n', 'none', 'h', 'hourly']:
#            sh_interval = int(sh_interval)
#            time.strptime(sh_time, '%H:%M')
#    except:
#        lg.warn('incorrect shedule '+s)
#        lg.exc()
#        return None
#    if sh_type in labels().keys():
#        sh_type = labels()[sh_type]
#    if sh_type not in labels().values():
#        lg.warn('incorrect shedule '+s)
#        return None
#    sh_details_new = ''
#    for i in range(len(sh_details)/3):
#        label = sh_details[i*3:i*3+3]
#        if sh_type == 'weekly' and not label in calendar.day_abbr:
#            lg.warn('incorrect shedule '+s)
#            return None
#        if sh_type == 'monthly' and not label in calendar.month_abbr:
#            lg.warn('incorrect shedule '+s)
#            return None
#        sh_details_new += label + ' '
#    return {'type': sh_type,
#            'interval': str(sh_interval),
#            'time': str(sh_time),
#            'details': sh_details_new.strip(),}


# def split(t):
#    typ = str(t[0])
#    if typ in types().keys():
#        typ = types()[typ]
#    if typ not in types().values():
#        lg.warn('incorrect shedule '+str(t))
#        return default()
#    return {
#        'type':         typ,
#        'time':         str(t[1]),
#        'interval':     str(t[2]),
#        'details':      str(t[3]), }


#-------------------------------------------------------------------------------


# def main():
#    nt = maths.shedule_next_daily(time.time()-60*60*24*2, 4, '12:00')
#    print 'daily', time.asctime(time.localtime(nt))
#    nw = maths.shedule_next_weekly(time.time(),1,'12:00', [0,])
#    print 'weekly', time.asctime(time.localtime(nw))
#    nm = maths.shedule_next_monthly(time.time(),1,'12:00', [1,])
#    print 'monthly', time.asctime(time.localtime(nm))
#    print unpack('weekly.4.18:45.MonTueWedThuFriSatSun')
#    print unpack('h.1')


if __name__ == '__main__':
    lg.set_debug_level(12)
    # main()



#!/usr/bin/python
# service_supplier_contracts.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (service_supplier_contracts.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#

"""
..

module:: service_supplier_contracts
"""

from __future__ import absolute_import
from services.local_service import LocalService


def create_service():
    return SupplierContractsService()


class SupplierContractsService(LocalService):

    service_name = 'service_supplier_contracts'
    config_path = 'services/supplier-contracts/enabled'

    def dependent_on(self):
        return [
            'service_supplier',
            'service_contract_chain',
        ]

    def installed(self):
        # TODO: to be continue...
        return False

    def start(self):
        from main import events
        from contacts import contactsdb
        from coins import supplier_contract_executor
        for customer_idurl in contactsdb.customers():
            supplier_contract_executor.init_contract(customer_idurl)
        events.add_subscriber(self._on_customer_modified, 'customer-modified')
        return True

    def stop(self):
        from main import events
        from coins import supplier_contract_executor
        events.remove_subscriber(self._on_customer_modified, 'customer-modified')
        for customer_idurl in list(supplier_contract_executor.all_contracts.keys()):
            supplier_contract_executor.shutdown_contract(customer_idurl)
        return True

    def _on_customer_modified(self, evt):
        from coins import supplier_contract_executor
        if evt.data.get('idurl'):
            supplier_contract_executor.recheck_contract(evt.data['idurl'])



import six
import time
import os

from unittest import TestCase


if six.PY2:
    from CodernityDB.database import Database
    from CodernityDB.hash_index import HashIndex
    from CodernityDB.tree_index import TreeBasedIndex

else:
    from CodernityDB3.database import Database
    from CodernityDB3.hash_index import HashIndex
    from CodernityDB3.tree_index import TreeBasedIndex


class WithXTreeIndex(TreeBasedIndex):

    def __init__(self, *args, **kwargs):
        kwargs['node_capacity'] = 10
        kwargs['key_format'] = 'I'
        super(WithXTreeIndex, self).__init__(*args, **kwargs)

    def make_key_value(self, data):
        t_val = data.get('x')
        if t_val is not None:
            return t_val, None
        return None

    def make_key(self, key):
        return key


class WithXHashIndex(HashIndex):

    def __init__(self, *args, **kwargs):
        kwargs['key_format'] = 'I'
        super(WithXHashIndex, self).__init__(*args, **kwargs)

    def make_key_value(self, data):
        a_val = data.get("x")
        if a_val is not None:
            return a_val, None
        return None

    def make_key(self, key):
        return key


def create_test_db(db_name='codernity_test_db_0', with_x_hash_index=False, with_x_tree_index=False):
    t = time.time()
    db_path = '/tmp/%s' % db_name
    os.system('rm -rf %s' % db_path)
    db = Database(db_path)
    db.create()
    if with_x_hash_index:
        x_ind = WithXHashIndex(db.path, 'x')
        db.add_index(x_ind)
    if with_x_tree_index:
        x_ind = WithXTreeIndex(db.path, 'x')
        db.add_index(x_ind)
    print('\ncreate_test_db finished in %f sec' % (time.time() - t))
    return db


class TestCodernityDB(TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_insert_save_store(self):
        _db = create_test_db(db_name='codernity_test_db_1', with_x_hash_index=True)
        for x in range(100):
            assert _db.insert(dict(x=x)) is not None
        x = 0
        for curr in _db.all('id'):
            assert curr['x'] == x
            x += 1

    def test_get_query(self):
        _db = create_test_db(db_name='codernity_test_db_2', with_x_hash_index=True)
        for x in range(100):
            _db.insert(dict(x=x))
        for y in range(100):
            _db.insert(dict(y=y))
        assert _db.get('x', 10, with_doc=True)['doc']['x'] == 10

    def test_deduplicates(self):
        _db = create_test_db(db_name='codernity_test_db_3', with_x_hash_index=True)
        for x in range(100):
            _db.insert(dict(x=x))
        for x in range(100):
            _db.insert(dict(x=x))
        for y in range(100):
            _db.insert(dict(y=y))
        assert _db.get('x', 10, with_doc=True)['doc']['x'] == 10
        assert len([i for i in _db.get_many('x', 10, limit=-1, with_doc=True)]) == 2

    def test_update_delete(self):
        _db = create_test_db(db_name='codernity_test_db_4', with_x_tree_index=True)
        for x in range(100):
            _db.insert(dict(x=x))
        for y in range(100):
            _db.insert(dict(y=y))
        assert _db.count(_db.all, 'x') == 100
        for curr in _db.all('x', with_doc=True):
            doc = curr['doc']
            if curr['key'] % 9 == 0:
                _db.delete(doc)
            elif curr['key'] % 5 == 0:
                doc['updated'] = True
                _db.update(doc)
        _db.reindex()
        total = _db.count(_db.all, 'x')
        assert total == 88

    def test_ordered(self):
        _db = create_test_db(db_name='codernity_test_db_5', with_x_tree_index=True)
        for x in range(11):
            _db.insert(dict(x=x))
        for y in range(11):
            _db.insert(dict(y=y))
        assert _db.get('x', 10, with_doc=True)['doc']['x'] == 10
        for curr in _db.get_many('x', start=3, end=8, limit=-1, with_doc=True):
            assert curr['doc']['x'] >= 3
